//
//  CVUAction.swift
//  MemriDatabase
//
//  Created by T Brennan on 8/1/21.
//

import Foundation
import UIKit
import SwiftUI

protocol CVUAction {
    init(vars: [String: CVUValue])
    func execute(sceneController: SceneController, context: CVUContext)
}

/// Used to look up the concrete type matching a CVU action name
func cvuAction(named: String) -> CVUAction.Type? {
    switch named.lowercased() {
    case "openview":
        return CVUAction_OpenView.self
    case "back":
        return CVUAction_NavigateBack.self
    case "forward":
        return CVUAction_NavigateBack.self
    case "additem":
        return CVUAction_AddItem.self
    case "copytoclipboard":
        return CVUAction_CopyToClipboard.self
    case "toggleeditmode":
        return CVUAction_ToggleEditMode.self
    case "togglefilterpanel":
        return CVUAction_ToggleFilterPanel.self
    case "togglenavigation":
        return CVUAction_ToggleNavigation.self
    case "togglefullscreen":
        return CVUAction_ToggleFullScreen.self
    case "delete":
        return CVUAction_Delete.self
    case "selectall":
        return CVUAction_SelectAll.self
    case "deselectall":
        return CVUAction_DeselectAll.self
    default:
        return nil
    }
}

struct CVUAction_OpenView: CVUAction {
    var vars: [String: CVUValue] = [:]
    
    var viewName: String?
    var renderer: String?
    var uids: Set<StringUUID>?
    var dateRange: DateInterval?
    
    func execute(sceneController: SceneController, context: CVUContext) {
        var viewDefinition: CVUDefinitionContent?
        if case let .subdefinition(viewDef) = vars["view"] {
            viewDefinition = viewDef
        }
        let viewArguments: CVUViewArguments
        if case let .subdefinition(viewArgs) = vars["viewArguments"] {
            viewArguments = CVUViewArguments(args: viewArgs.properties, argumentItem: context.currentItem, parentArguments: context.viewArguments)
        } else {
            viewArguments = CVUViewArguments()
        }
        let db = sceneController.appController.databaseController
        let resolver = CVUPropertyResolver(context: context, lookup: CVULookupController(), db: db, properties: vars)
        sceneController.navigateToNewContext(viewName: viewName ?? resolver.string("viewName") ?? "customView",
                                             inheritDatasource: resolver.bool("inheritDatasource", defaultValue: true),
                                             overrideRenderer: renderer ?? resolver.string("renderer"),
                                             defaultRenderer: "singleItem",
                                             targetItem: context.currentItem,
                                             overrideUIDs: uids,
                                             dateRange: dateRange, customDefinition: viewDefinition,
                                             viewArguments: viewArguments)
    }
}

extension CVUAction_OpenView {
    init(vars: [String: CVUValue]) {
        self.vars = vars
    }
}

struct CVUAction_NavigateBack: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext) {
        sceneController.navigateBack()
    }
}

struct CVUAction_CopyToClipboard: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext) {
        let db = sceneController.appController.databaseController
        let resolver = CVUPropertyResolver(context: context, lookup: CVULookupController(), db: db, properties: vars)
        if let value = resolver.string("value") {
            UIPasteboard.general.string = value
        }
    }
}

struct CVUAction_AddItem: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext) {
        let db = sceneController.appController.databaseController
        let resolver = CVUPropertyResolver(context: context, lookup: CVULookupController(), db: db, properties: vars)
        if let template = resolver.subdefinition("template"), let type = template.string("_type") {
            let item = ItemRecord(type: type)
            
            /// Take all the properties defined in the template definition (in CVU) and map them against the schema. Resolve the CVU based on the type expected by the schema.
            let propertiesAndEdges = template.properties.keys.compactMap { key -> BaseRecord? in
                guard let valueType = db.schema.expectedType(forItemType: type, propertyOrEdgeName: key) else { return nil }
                switch valueType {
                case .property(let propertyType):
                    guard let value: PropertyDatabaseValue = ({
                        switch propertyType {
                        case .string:
                            return template.string(key).map { .string($0) }
                        case .bool:
                            return template.bool(key).map { .bool($0) }
                        case .int:
                            return template.int(key).map { .int($0) }
                        case .double:
                            return template.double(key).map { .double($0) }
                        case .datetime:
                            return nil
                        case .blob:
                            return nil
                        }
                    }()) else { return nil }
                    return ItemPropertyRecord(itemUID: item.uid, name: key, value: value)
                case .edge:
                    guard let target = template.item(key) else { return nil }
                    return ItemEdgeRecord(sourceUID: item.uid, name: key, targetUID: target.uid)
                }
            }
            db.write { (db) in
                try item.save(db)
                try propertiesAndEdges.forEach { try $0.save(db) }
            }
            CVUAction_OpenView(vars: vars).execute(sceneController: sceneController, context: context.replacingItem(item))
        }
    }
}


struct CVUAction_ToggleEditMode: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext) {
        //
    }
}

struct CVUAction_ToggleFilterPanel: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext) {
        //
    }
}


struct CVUAction_ToggleNavigation: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext) {
        withAnimation {
            sceneController.navigationIsVisible.toggle()
        }
    }
}

struct CVUAction_ToggleFullScreen: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext) {
        //
    }
}

struct CVUAction_Delete: CVUAction {
    var vars: [String: CVUValue] = [:]
    
    func execute(sceneController: SceneController, context: CVUContext) {
        guard let item = context.currentItem else { return }
        sceneController.appController.databaseController.write { (db) in
            try item.delete(db)
        }
    }
}


struct CVUAction_SelectAll: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext) {
        //
    }
}

struct CVUAction_DeselectAll: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext) {
        //
    }
}
