//
// CVUParsedDefinition.swift
// Copyright © 2020 memri. All rights reserved.

import Foundation

enum CVUDefinitionType: String, Codable {
    case view
    case views
    case uiNode
    case sessions
    case renderer
    case datasource
    case language
    case other
}

enum CVUDefinitionDomain: String, Codable {
    case user
}

struct CVUDefinitionContent: Equatable, CVUStringConvertible, Codable {
    var definitions: [CVUParsedDefinition] = []
    var children: [CVUUINode] = []
    var properties: [String: CVUValue] = [:]
    
    func propertyResolver(context: CVUContext, lookup: CVULookupController, db: DatabaseController) -> CVUPropertyResolver {
        CVUPropertyResolver(context: context, lookup: lookup, db: db, properties: properties)
    }
    
    func toCVUString(depth: Int, tab: String, includeInitialTab: Bool) -> String {
        let tabs = Array(repeating: tab, count: depth).joined()
        let propertiesString = properties.isEmpty ? "" : properties.toCVUString(depth: depth + 1, tab: tab, includeInitialTab: true).appending("\n")
        
        let childrenString = children.isEmpty ? "" : children.map { node in
            node.toCVUString(depth: depth + 1, tab: tab, includeInitialTab: true)
        }.joined(separator: "\n").appending("\n")
        
        let nestedDefinitions = definitions.isEmpty ? "" : definitions.map {
            $0.toCVUString(depth: depth + 1, tab: tab, includeInitialTab: true)
        }.joined(separator: "\n\n").appending("\n")
        
        let newLineA = !propertiesString.isEmpty && !childrenString.isEmpty
        let newLineB = !nestedDefinitions.isEmpty && (!propertiesString.isEmpty || !childrenString.isEmpty)
        
        return "\(includeInitialTab ? tabs : ""){\n\(propertiesString)\(newLineA ? "\n" : "")\(childrenString)\(newLineB ? "\n" : "")\(nestedDefinitions)\(tabs)}"
    }
    
    func merge(with other: CVUDefinitionContent?) -> CVUDefinitionContent {
        guard let other = other else { return self }
        var result = self
        for definition in other.definitions {
            if let index = result.definitions.firstIndex(where: { $0.selector == definition.selector }) {
                result.definitions[index] = result.definitions[index].merge(with: definition)
            } else {
                result.definitions.append(definition)
            }
        }
        result.properties.merge(other.properties, uniquingKeysWith: { lhs, rhs in
            if let lhs = lhs.getSubdefinition(), let rhs = rhs.getSubdefinition() {
                return .subdefinition(lhs.merge(with: rhs))
            } else {
                return rhs // Prefer rhs properties
            }
        })
        if !other.children.isEmpty {
            result.children = other.children // Override children if defined
        }
        
        return result
    }
}

public struct CVUParsedDefinition: Equatable, CVUStringConvertible, Codable {
    var type: CVUDefinitionType = .other
    var domain: CVUDefinitionDomain = .user
    var selector: String?
    var name: String?
    // The renderer for which this definition is valid
    var renderer: String?
    
    var parsed: CVUDefinitionContent = CVUDefinitionContent()
    
    subscript(propName: String) -> CVUValue? {
        get {
            parsed.properties[propName].flatMap { $0 }
        }
        set(value) {
            parsed.properties[propName] = value
        }
    }

    var selectorIsForList: Bool {
        selector?.hasSuffix("[]") ?? false
    }
    
    var description: String {
        toCVUString(depth: 0, tab: "   ", includeInitialTab: true)
    }

    func toCVUString(depth: Int, tab: String, includeInitialTab: Bool) -> String {
        let tabs = Array(repeating: tab, count: depth).joined()
        let body = parsed.toCVUString(depth: depth, tab: tab, includeInitialTab: false)
        return "\(includeInitialTab ? tabs : "")\(selector?.nilIfBlank.map { "\($0) "} ?? "")\(renderer?.nilIfBlank.map { "> \($0) "} ?? "")\(body)"
    }
    
    func merge(with other: CVUParsedDefinition?) -> CVUParsedDefinition {
        guard let other = other else { return self }
        var result = self
        result.parsed = parsed.merge(with: other.parsed)
        return result
    }
}
