//
//  CVUValue_Expression.swift
//  Memri
//
//  Created by T Brennan on 28/1/21.
//

import Foundation

extension CVUValue {
    indirect enum ExpressionNode: Equatable, AutoCodable {
        typealias ExpressionNode = CVUValue.ExpressionNode
        typealias LookupNode = CVUValue.LookupNode
        typealias Constant = CVUValue.Constant
        
        case lookup(nodes: [LookupNode])
        case stringMode(nodes: [ExpressionNode])
        case conditional(condition: ExpressionNode, trueExp: ExpressionNode, falseExp: ExpressionNode)
        case or(ExpressionNode, ExpressionNode)
        case negation(ExpressionNode)
        case addition(ExpressionNode, ExpressionNode)
        case subtraction(ExpressionNode, ExpressionNode)
        case multiplication(ExpressionNode, ExpressionNode)
        case division(ExpressionNode, ExpressionNode)
        case constant(Constant)
        case lessThan(ExpressionNode, ExpressionNode)
        case greaterThan(ExpressionNode, ExpressionNode)
        case lessThanOrEqual(ExpressionNode, ExpressionNode)
        case greaterThanOrEqual(ExpressionNode, ExpressionNode)
        case areEqual(ExpressionNode, ExpressionNode)
        case areNotEqual(ExpressionNode, ExpressionNode)
        case and(ExpressionNode, ExpressionNode)
        
        var description: String {
            "\(self)"
        }
        
        
    }
}

extension CVUValue.ExpressionNode {
    func toCVUString() -> String {
        switch self {
        case .lookup(let nodes):
            return nodes.map { $0.toCVUString() }.joined(separator: ".")
        case .stringMode(nodes: let nodes):
            return nodes.map {
                switch $0 {
                case .constant(let constant):
                    return constant.toCVUString(insideStringMode: true)
                default:
                    return "{\($0.toCVUString())}"
                }
            }.joined()
        case let .conditional(condition, trueExp, falseExp):
            return "\(condition.toCVUString()) ? \(trueExp.toCVUString()) : \(falseExp.toCVUString())"
        case let .or(lhs, rhs):
            return "\(lhs.toCVUString()) OR \(rhs.toCVUString())"
        case .negation(let x):
            return "!\(x.toCVUString())"
        case let .addition(lhs, rhs):
            return "\(lhs.toCVUString()) + \(rhs.toCVUString())"
        case let .subtraction(lhs, rhs):
            return "\(lhs.toCVUString()) - \(rhs.toCVUString())"
        case let .multiplication(lhs, rhs):
            return "\(lhs.toCVUString()) * \(rhs.toCVUString())"
        case let .division(lhs, rhs):
            return "\(lhs.toCVUString()) / \(rhs.toCVUString())"
        case .constant(let x):
            return x.toCVUString()
        case let .lessThan(lhs, rhs):
            return "\(lhs.toCVUString()) < \(rhs.toCVUString())"
        case let .greaterThan(lhs, rhs):
            return "\(lhs.toCVUString()) > \(rhs.toCVUString())"
        case let .lessThanOrEqual(lhs, rhs):
            return "\(lhs.toCVUString()) <= \(rhs.toCVUString())"
        case let .greaterThanOrEqual(lhs, rhs):
            return "\(lhs.toCVUString()) >= \(rhs.toCVUString())"
        case let .areEqual(lhs, rhs):
            return "\(lhs.toCVUString()) = \(rhs.toCVUString())"
        case let .areNotEqual(lhs, rhs):
            return "\(lhs.toCVUString()) != \(rhs.toCVUString())"
        case let .and(lhs, rhs):
            return "\(lhs.toCVUString()) AND \(rhs.toCVUString())"
        }
    }
}
