//
//  CVUExpressionParser.swift
//  MemriDatabase
//
//  Created by T Brennan on 7/12/20.
//

import Foundation

enum CVUExpressionParseErrors: Error {
    case UnexpectedToken(CVUExpressionLexer.ExprToken)
    case UndefinedOperator(String)

    case ExpectedCharacter(Character)
    case ExpectedExpression(CVUExpressionLexer.ExprToken)
    case ExpectedArgumentList
    case ExpectedIdentifier
    case ExpectedConditionElse

    case MissingQuoteClose
}

class CVUExpressionParser {
    let tokens: [CVUExpressionLexer.ExprToken]
    var index = 0
    var lastToken: CVUExpressionLexer.ExprToken?
    var countStringModeNodes = 0

    typealias ExpressionNode = CVUValue.ExpressionNode
    typealias ExprToken = CVUExpressionLexer.ExprToken

    init(_ tokens: [ExprToken]) {
        self.tokens = tokens
    }

    func peekCurrentToken() -> CVUExpressionLexer.ExprToken {
        index >= tokens.count
            ? ExprToken.EOF
            : tokens[index]
    }

    func popCurrentToken() -> CVUExpressionLexer.ExprToken {
        if index >= tokens.count {
            lastToken = CVUExpressionLexer.ExprToken.EOF
            return CVUExpressionLexer.ExprToken.EOF
        }

        lastToken = tokens[index]
        index += 1
        return lastToken ?? ExprToken.EOF // Check for out of bound?
    }

    func parse() throws -> CVUValue.ExpressionNode {
        index = 0
        let result = try parseExpression()

        if case ExprToken.EOF = popCurrentToken() {
            return result
        }
        throw CVUExpressionParseErrors.UnexpectedToken(lastToken!)
    }

    func parseExpression() throws -> CVUValue.ExpressionNode {
        let node = try parsePrimary()
        return try parseBinaryOp(node)
    }

    func parsePrimary(_ skipOperator: Bool = false) throws -> CVUValue.ExpressionNode {
        switch peekCurrentToken() {
        case .Negation:
            return try parseNegation()
        case .Identifier:
            return try parseIdentifier()
        case .Number:
            return try parseNumber()
        case .String:
            return try parseString()
        case .Bool:
            return try parseBool()
        case .CurlyBracketOpen:
            return try parseCurlyBrackets()
        case .ParensOpen:
            return try parseParens()
        case .Period:
            return try parsePeriod()
        case .Operator:
            if skipOperator { fallthrough }
            else { return try parseOperator() } // start with + or -
        default:
            throw CVUExpressionParseErrors.ExpectedExpression(popCurrentToken())
        }
    }

    func parseLookupExpression() throws -> CVUValue.ExpressionNode {
        try parseExpression() // TODO: maybe: This could be limited to int and string
    }

    func parseIntExpressionComponent() throws -> CVUValue.ExpressionNode {
        try parsePrimary(true)
    }

    func parseNumber() throws -> CVUValue.ExpressionNode {
        guard case let ExprToken.Number(value, _) = popCurrentToken() else {
            throw CVUExpressionParseErrors.UnexpectedToken(lastToken!)
        }
        return .constant(.number(value))
    }

    func parseString() throws -> CVUValue.ExpressionNode {
        guard case let ExprToken.String(value, _) = popCurrentToken() else {
            throw CVUExpressionParseErrors.UnexpectedToken(lastToken!)
        }
        return .constant(.string(value))
    }

    func parseBool() throws -> ExpressionNode {
        guard case let ExprToken.Bool(value, _) = popCurrentToken() else {
            throw CVUExpressionParseErrors.UnexpectedToken(lastToken!)
        }
        return .constant(.bool(value))
    }

    func parsePeriod() throws -> ExpressionNode {
        guard case ExprToken.Period = peekCurrentToken() else {
            throw CVUExpressionParseErrors.UnexpectedToken(lastToken!)
        }
        return try parseIdentifier(initialNode: .default)
    }

    func parseOperator() throws -> ExpressionNode {
        guard case let ExprToken.Operator(op, _) = popCurrentToken() else {
            throw CVUExpressionParseErrors.UnexpectedToken(lastToken!)
        }
        if op == .Minus {
            let exp = try parseIntExpressionComponent()
            return .subtraction(
                .constant(.number(0)),
                exp
            )
        }
        else if op == .Plus {
            let exp = try parseIntExpressionComponent()
            return exp
        }
        else {
            throw CVUExpressionParseErrors.UnexpectedToken(lastToken!)
        }
    }

    func parseNegation() throws -> ExpressionNode {
        guard case ExprToken.Negation = popCurrentToken() else {
            throw CVUExpressionParseErrors.UnexpectedToken(lastToken!)
        }

        let exp = try parsePrimary()

        return .negation(exp)
    }

    func parseCurlyBrackets() throws -> ExpressionNode {
        guard case ExprToken.CurlyBracketOpen = popCurrentToken() else {
            throw CVUExpressionParseErrors.ExpectedCharacter("{")
        }

        return try parseStringMode()
    }

    func parseParens() throws -> ExpressionNode {
        guard case ExprToken.ParensOpen = popCurrentToken() else {
            throw CVUExpressionParseErrors.ExpectedCharacter("(")
        }

        let exp = try parseExpression()

        guard case ExprToken.ParensClose = popCurrentToken() else {
            throw CVUExpressionParseErrors.ExpectedCharacter(")")
        }

        return exp
    }

    func parseIdentifier(initialNode: CVUValue.LookupNode? = nil) throws -> ExpressionNode {
        var sequence: [CVUValue.LookupNode] = []

        if let initialNode = initialNode {
            sequence.append(initialNode)
        }

        while true {
            if case let ExprToken.Identifier(name, _) = peekCurrentToken() {
                _ = popCurrentToken()
                sequence.append(CVUValue.LookupNode(name: name, type: .lookup()))
            }
            
            if case ExprToken.BracketOpen = peekCurrentToken() {
                _ = popCurrentToken()

                guard !sequence.isEmpty else {
                    throw CVUExpressionParseErrors.ExpectedIdentifier
                }

                if case ExprToken.BracketClose = peekCurrentToken() {
                    _ = popCurrentToken()
                    sequence[sequence.endIndex - 1].isArray = true
                }
                else {
                    let exp = try parseLookupExpression()
                    sequence[sequence.endIndex - 1].type = .lookup(subExpression: exp)

                    guard case ExprToken.BracketClose = popCurrentToken() else {
                        throw CVUExpressionParseErrors.ExpectedCharacter("]")
                    }
                }
            }
            
            
            if case ExprToken.ParensOpen = peekCurrentToken() {
                _ = popCurrentToken()
                var arguments: [ExpressionNode] = []
                if case ExprToken.ParensClose = peekCurrentToken() {
                    _ = popCurrentToken()
                }
                else {
                    ArgumentsLoop: while true {
                        let argument = try parseExpression()
                        arguments.append(argument)

                        if case ExprToken.ParensClose = peekCurrentToken() {
                            _ = popCurrentToken()
                            break ArgumentsLoop
                        }

                        guard case ExprToken.Comma = popCurrentToken() else {
                            throw CVUExpressionParseErrors.ExpectedArgumentList
                        }
                    }
                }
                
                guard !sequence.isEmpty else {
                    throw CVUExpressionParseErrors.ExpectedIdentifier
                }
                sequence[sequence.endIndex - 1].type = .function(args: arguments)
            }

            
            let nextToken = peekCurrentToken()
            
            if case ExprToken.Period = nextToken {
                _ = popCurrentToken()
                continue
            } else {
                break
            }
        }

        return ExpressionNode.lookup(nodes: sequence)
    }

    func getCurrentTokenPrecedence() throws -> Int {
        guard index < tokens.count else { return -1 }

        let nextToken = peekCurrentToken()
        guard case let ExprToken.Operator(op, _) = nextToken else {
            if case ExprToken.CurlyBracketOpen = nextToken { return 1 }
            if case ExprToken.CurlyBracketClose = nextToken { return 2 }

            return -1
        }

        return op.precedence
    }

    func parseBinaryOp(_ node: ExpressionNode, exprPrecedence: Int = 0) throws -> ExpressionNode {
        var lhs = node
        while true {
            let tokenPrecedence = try getCurrentTokenPrecedence()
            if tokenPrecedence < exprPrecedence {
                return lhs
            }

            let nextToken = peekCurrentToken()
            if case let ExprToken.Operator(op, _) = nextToken, op == .ConditionElse { return lhs }
            if case ExprToken.CurlyBracketClose = nextToken { return lhs }

            guard case let ExprToken.Operator(op, _) = popCurrentToken() else {
                if case ExprToken.CurlyBracketOpen = lastToken! {
                    return try parseStringMode(lhs)
                }

                throw CVUExpressionParseErrors.UnexpectedToken(lastToken!)
            }

            if op == .ConditionStart { return try parseConditionOp(lhs) }

            var rhs = try parsePrimary()
            let nextPrecedence = try getCurrentTokenPrecedence()

            if tokenPrecedence < nextPrecedence {
                rhs = try parseBinaryOp(rhs, exprPrecedence: tokenPrecedence + 1)
            }
            
            switch op {
            case .ConditionEquals:
                lhs = .areEqual(lhs, rhs)
            case .ConditionNotEquals:
                lhs = .areNotEqual(lhs, rhs)
            case .Plus:
                lhs = .addition(lhs, rhs)
            case .Minus:
                lhs = .subtraction(lhs, rhs)
            case .Multiplication:
                lhs = .multiplication(lhs, rhs)
            case .Division:
                lhs = .division(lhs, rhs)
            case .ConditionOR:
                lhs = .or(lhs, rhs)
            case .ConditionGreaterThan:
                lhs = .greaterThan(lhs, rhs)
            case .ConditionGreaterThanOrEqual:
                lhs = .greaterThanOrEqual(lhs, rhs)
            case .ConditionLessThan:
                lhs = .lessThan(lhs, rhs)
            case .ConditionLessThanOrEqual:
                lhs = .lessThanOrEqual(lhs, rhs)
            case .ConditionAND:
                lhs = .and(lhs, rhs)
            default:
                throw CVUExpressionParseErrors.UndefinedOperator(op.rawValue)
            }
        }
    }

    func parseConditionOp(_ conditionNode: ExpressionNode) throws -> ExpressionNode {
        let trueExp = try parseExpression()

        guard case let ExprToken.Operator(op, _) = popCurrentToken() else {
            throw CVUExpressionParseErrors.ExpectedConditionElse
        }

        if op != .ConditionElse {
            throw CVUExpressionParseErrors.ExpectedConditionElse
        }

        let falseExp = try parseExpression()

        return .conditional(condition: conditionNode, trueExp: trueExp, falseExp: falseExp)
    }

    func parseStringMode(_ firstNode: ExpressionNode? = nil) throws -> ExpressionNode {
        countStringModeNodes += 1
        if countStringModeNodes > 1 {
            throw CVUExpressionParseErrors.UnexpectedToken(lastToken!)
        }

        var expressions = [ExpressionNode]()
        if let firstNode = firstNode { expressions.append(firstNode) }

        while true {
            let nextToken = peekCurrentToken()
            if case ExprToken.EOF = nextToken { break }
            if case ExprToken.String = nextToken {
                expressions.append(try parseString())
                continue
            }
            if case ExprToken.CurlyBracketOpen = nextToken {
                _ = popCurrentToken()
            }

            expressions.append(try parseExpression())

            guard case ExprToken.CurlyBracketClose = popCurrentToken() else {
                if case ExprToken.EOF = lastToken! { break }
                throw CVUExpressionParseErrors.ExpectedCharacter("}")
            }
        }

        return .stringMode(nodes: expressions)
    }
}
