//
//  CVULookup.swift
//  MemriDatabase
//
//  Created by T Brennan on 23/12/20.
//

import Foundation

struct CVULookupController {
    init(mockMode: LookupMock? = nil) {
        self.lookupMockMode = mockMode
    }
    
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> Double? {
        switch value {
        case .constant(let constant):
            return constant.asNumber()
        case .expression(let expression):
            return resolve(expression: expression, context: context, db: db)
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> String? {
        switch value {
        case .constant(let constant):
            return constant.asString()
        case .expression(let expression):
            return resolve(expression: expression, context: context, db: db)
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> Bool? {
        switch value {
        case .constant(let constant):
            return constant.asBool()
        case .expression(let expression):
            return resolve(expression: expression, context: context, db: db)
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> Date? {
        switch value {
        case .constant(let constant):
            return constant.asNumber().flatMap { DatabaseHelperFunctions.decode(from: Int($0)) }
        case .expression(let expression):
            return resolve(expression: expression, context: context, type: Double.self, db: db).flatMap { DatabaseHelperFunctions.decode(from: Int($0)) }
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> [String: CVUValue]? {
        switch value {
        case .dictionary(let dict):
            return dict
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> ItemRecord? {
        switch value {
        case .constant:
            return nil
        case .expression(let expression):
            return resolve(expression: expression, context: context, db: db)
        default:
            return nil
        }
    }
    
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> [ItemRecord] {
        switch value {
        case .expression(let expression):
            return resolve(expression: expression, context: context, db: db)
        default:
            return []
        }
    }
    
    func resolve(value: CVUValue, defaultValue: Bool, context: CVUContext, db: DatabaseController) -> Binding<Bool>? {
        switch value {
        case let .expression(.lookup(nodes)):
            guard let (item, property) = resolveToItemAndProperty(nodes: nodes, context: context, db: db)
            else { return nil }
            return item.propertyBinding(name: property, defaultValue: defaultValue)
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, defaultValue: String, context: CVUContext, db: DatabaseController) -> Binding<String>? {
        switch value {
        case let .expression(.lookup(nodes)):
            guard let (item, property) = resolveToItemAndProperty(nodes: nodes, context: context, db: db)
            else { return nil }
            return item.propertyBinding(name: property, defaultValue: defaultValue)
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, defaultValue: String?, context: CVUContext, db: DatabaseController) -> Binding<String?>? {
        switch value {
        case let .expression(.lookup(nodes)):
            guard let (item, property) = resolveToItemAndProperty(nodes: nodes, context: context, db: db)
            else { return nil }
            return item.propertyBinding(name: property, defaultValue: defaultValue)
        default:
            return nil
        }
    }
    
    /// Lookup an edge from an ItemRecord
    func resolve(edge: String, for item: ItemRecord, db: DatabaseController) -> ItemRecord? {
        item.edgeItem(edge, db: db)
    }
    
    /// Lookup an edge array from an ItemRecord
    func resolve(edge: String, for item: ItemRecord, db: DatabaseController) -> [ItemRecord] {
        item.edgeItems(edge, db: db)
    }
    
    /// Lookup a property from an ItemRecord
    func resolve(property: String, for item: ItemRecord, db: DatabaseController) -> PropertyDatabaseValue? {
        item.property(property, db: db)?.value(itemType: item.type, schema: db.schema)
    }
    
    
    
    
    
    
    
    struct LookupMock {
        var bool: Bool
        var string: String
        var double: Double
    }
    
    let lookupMockMode: LookupMock?
    
    enum LookupStep {
        case items([ItemRecord])
        case values([PropertyDatabaseValue])
    }
    
    func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> LookupStep? {
        var currentValue: LookupStep?
        for node in nodes {
            switch node.type {
            case .default:
                guard let currentItem = context.currentItem else { return nil }
                currentValue = .items([currentItem])
            case .function(let args):
                switch node.name.lowercased() {
                case "joined":
                    guard case let .values(values) = currentValue else { return nil }
                    if let exp = args.first, let separator = resolve(expression: exp, context: context, type: String.self, db: db) {
                        let joined = values.compactMap { $0.asString()?.nilIfBlank }.joined(separator: separator)
                        currentValue = .values([.string(joined)])
                    } else {
                        let strings = values.compactMap { $0.asString()?.nilIfBlank }
                        let joined = ListFormatter.localizedString(byJoining: strings)
                        currentValue = .values([.string(joined)])
                    }
                case "joinwithcomma":
                    let strings = args.compactMap {
                        resolve(expression: $0, context: context, type: String.self, db: db)?.nilIfBlank
                    }
                    let joined = strings.joined(separator: ", ")
                    currentValue = .values([.string(joined)])
                case "joinnatural":
                    let strings = args.compactMap {
                        resolve(expression: $0, context: context, type: String.self, db: db)?.nilIfBlank
                    }
                    let joined = ListFormatter.localizedString(byJoining: strings)
                    currentValue = .values([.string(joined)])
                case "plainstring":
                    guard case let .values(values) = currentValue else { return nil }
                    let stripped = values.compactMap { value -> PropertyDatabaseValue? in
                        guard let htmlstring = value.asString() else { return nil }
                        return .string(HTMLHelper.getPlainText(html: htmlstring))
                    }
                    currentValue = .values(stripped)
                case "first":
                    if case let .values(values) = currentValue, let first = values.first { currentValue = .values([first]) }
                    else if case let .items(items) = currentValue, let first = items.first { currentValue = .items([first]) }
                    else { return nil }
                case "last":
                    if case let .values(values) = currentValue, let last = values.last { currentValue = .values([last]) }
                    else if case let .items(items) = currentValue, let last = items.last { currentValue = .items([last]) }
                    else { return nil }
                case "fullname":
                    switch currentValue {
                    case let .items(items):
                        if let first = items.first, first.type == "Person" {
                            let name = [
                                resolve(property: "firstName", for: first, db: db),
                                resolve(property: "lastName", for: first, db: db)
                            ].compactMap { $0?.asString() }.joined(separator: " ")
                            currentValue = .values([.string(name)])
                        } else {
                            return nil
                        }
                    default:
                        return nil
                }
                case "initials":
                    switch currentValue {
                    case let .values(values):
                        let initials = String(values.compactMap { $0.asString()?.first }).uppercased()
                        currentValue = .values([.string(initials)])
                    case let .items(items):
                        if let first = items.first, first.type == "Person" {
                            let initials = String([
                                resolve(property: "firstName", for: first, db: db),
                                resolve(property: "lastName", for: first, db: db)
                            ].compactMap { $0?.asString()?.first }).uppercased()
                            currentValue = .values([.string(initials)])
                        } else {
                            return nil
                        }
                    case .none:
                        return nil
                }
                default: return nil
                }
            case .lookup(let subexpression):
                switch currentValue {
                case .items(let items):
                    currentValue = itemLookup(node: node, items: items, subexpression: subexpression, context: context, db: db)
                case .none:
                        // Check if there is a matching view argument
                        if let viewArgs = context.viewArguments, let argValue = viewArgs.args[node.name] {
                            switch argValue {
                            case .constant(let constant):
                                switch constant {
                                case let .argument(x):
                                    currentValue = .values([.string(x)])
                                case let .number(x):
                                    currentValue = .values([.double(x)])
                                case let .string(x):
                                    currentValue = .values([.string(x)])
                                case let .bool(x):
                                    currentValue = .values([.bool(x)])
                                case let .colorHex(x):
                                    currentValue = .values([.string(x)])
                                case .nil:
                                    currentValue = nil
                                }
                            case .expression(let expression):
                                let context = CVUContext(currentItem: viewArgs.argumentItem, viewArguments: viewArgs.parentArguments)
                                if let item: ItemRecord = resolve(expression: expression, context: context, db: db) {
                                    currentValue = .items([item])
                                } else if let number: Double = resolve(expression: expression, context: context, db: db) {
                                    currentValue = .values([.double(number)])
                                } else if let string: String = resolve(expression: expression, context: context, db: db) {
                                    currentValue = .values([.string(string)])
                                } else {
                                    let items: [ItemRecord] = resolve(expression: expression, context: context, db: db)
                                        if !items.isEmpty {
                                            currentValue = .items(items)
                                        } else {
                                            return nil
                                        }
                                }
                            default:
                                return nil
                            }
                        }
                default:
                    return nil
                }
            }
        }
        return currentValue
    }
    
    func filter(items: [ItemRecord], by subexpression: CVUValue.ExpressionNode?, db: DatabaseController) -> [ItemRecord] {
        guard let exp = subexpression else { return items }
        return items.filter { item in
            let context = CVUContext(currentItem: item)
            return resolve(expression: exp, context: context, db: db) ?? false
        }
    }

    func itemLookup(node: CVUValue.LookupNode, items: [ItemRecord], subexpression: CVUValue.ExpressionNode?, context: CVUContext, db: DatabaseController) -> LookupStep? {
        let trimmedName = node.name.trimmingCharacters(in: ["~"])
        switch node.name.first {
        case "~":
            /// LOOKUP REVERSE EDGE FOR EACH ITEM
            if node.isArray {
                let result = items.flatMap { (item) -> [ItemRecord] in
                    return item.reverseEdgeItems(trimmedName, db: db)
                }
                return .items(filter(items: result, by: subexpression, db: db))
            } else {
                let result = items.compactMap { (item) -> ItemRecord? in
                    return item.reverseEdgeItem(trimmedName, db: db)
                }
                return .items(filter(items: result, by: subexpression, db: db))
            }
        default:
            /// CHECK IF WE'RE EXPECTING EDGES OR PROPERTIES
            guard let itemType = items.first?.type else {
                return nil
            }
            /// Check if this is an intrinsic property
            switch node.name {
            case "uid":
                return .values(items.map { return .string($0.uid.uid) })
            case "dateModified":
                return .values(items.map { return .datetime($0.dateModified) })
            case "dateCreated":
                return .values(items.map { return .datetime($0.dateCreated) })
            default: break
            }
            /// Find out if it is a property or an edge (according to schema)
            switch db.schema.expectedType(forItemType: itemType, propertyOrEdgeName: node.name) {
            case .property:
                /// LOOKUP PROPERTY FOR EACH ITEM
                let result = items.compactMap { (item) -> PropertyDatabaseValue? in
                    guard let property = item.property(node.name, db: db),
                          let value = property.value(itemType: item.type, schema: db.schema)
                    else { return nil}
                    return value
                }
                return .values(result)
            case .edge:
                /// LOOKUP EDGE FOR EACH ITEM
                if node.isArray {
                    let result = items.flatMap { (item) -> [ItemRecord] in
                        return item.edgeItems(trimmedName, db: db)
                    }
                    return .items(filter(items: result, by: subexpression, db: db))
                } else {
                    let result = items.compactMap { (item) -> ItemRecord? in
                        return item.edgeItem(trimmedName, db: db)
                    }
                    return .items(filter(items: result, by: subexpression, db: db))
                }
            case .none:
                return nil
            }
        }
    }
    
    
    
    
    
    /// Lookup a variable using its CVU string and return the value as a double
    func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> Double? {
        if let mock = lookupMockMode { return mock.double }
        guard let lookupResult: LookupStep = resolve(nodes: nodes, context: context, db: db) else { return nil }
        switch lookupResult {
        case .values(let values):
            return values.first?.asDouble()
        default:
            return nil
        }
    }
    
    /// Lookup a variable using its CVU string and return the value as a string
    func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> String? {
        if let mock = lookupMockMode { return mock.string }
        guard let lookupResult: LookupStep = resolve(nodes: nodes, context: context, db: db) else { return nil }
        switch lookupResult {
        case .values(let values):
            return values.first?.asString()
        default:
            return nil
        }
    }
    
    /// Lookup a variable using its CVU string and return the value as a bool
    func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> Bool? {
        if let mock = lookupMockMode { return mock.bool }
        guard let lookupResult: LookupStep = resolve(nodes: nodes, context: context, db: db) else { return nil }
        switch lookupResult {
        case .values(let values):
            return values.first?.asBool()
        case .items(let items):
         return !items.isEmpty
        }
    }
    
    /// Lookup using a CVU expression string and return the value as an item
    func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> ItemRecord? {
        guard let lookupResult: LookupStep = resolve(nodes: nodes, context: context, db: db) else { return nil }
        switch lookupResult {
        case .items(let items):
            return items.first
        default:
            return nil
        }
    }
    
    /// Lookup using a CVU expression string and return the value as an array of items
    func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> [ItemRecord] {
        guard let lookupResult: LookupStep = resolve(nodes: nodes, context: context, db: db) else { return [] }
        switch lookupResult {
        case .items(let items):
            return items
        default:
            return []
        }
    }
    
    
    func resolve(expression: CVUValue.ExpressionNode, context: CVUContext, type: ItemRecord.Type = ItemRecord.self, db: DatabaseController) -> ItemRecord? {
        switch expression {
        case let .lookup(nodes):
            return resolve(nodes: nodes, context: context, db: db)
        case let .conditional(condition, trueExp, falseExp):
            let conditionResolved = resolve(expression: condition, context: context, type: Bool.self, db: db) ?? false
            if conditionResolved {
                return resolve(expression: trueExp, context: context, db: db)
            } else {
                return resolve(expression: falseExp, context: context, db: db)
            }
        case let .or(a, b):
            return resolve(expression: a, context: context, type: ItemRecord.self, db: db) ?? resolve(expression: b, context: context, db: db)
        default:
//            print("CVU Expression: \(expression) unsupported for resolving to ItemRecord")
            return nil
        }
    }
    
    func resolve(expression: CVUValue.ExpressionNode, context: CVUContext, type: [ItemRecord].Type = [ItemRecord].self, db: DatabaseController) -> [ItemRecord] {
        switch expression {
        case let .lookup(nodes):
            return resolve(nodes: nodes, context: context, db: db)
        case let .conditional(condition, trueExp, falseExp):
            let conditionResolved = resolve(expression: condition, context: context, type: Bool.self, db: db) ?? false
            if conditionResolved {
                return resolve(expression: trueExp, context: context, db: db)
            } else {
                return resolve(expression: falseExp, context: context, db: db)
            }
        case let .and(a, b):
            return resolve(expression: a, context: context, type: [ItemRecord].self, db: db) + resolve(expression: b, context: context, db: db)
        default:
//            print("CVU Expression: \(expression) unsupported for resolving to [ItemRecord]")
            return []
        }
    }
    
    func resolve(expression: CVUValue.ExpressionNode, context: CVUContext, type: Double.Type = Double.self, db: DatabaseController) -> Double? {
        switch expression {
        case let .lookup(nodes):
            return resolve(nodes: nodes, context: context, db: db)
        case let .conditional(condition, trueExp, falseExp):
            let conditionResolved = resolve(expression: condition, context: context, type: Bool.self, db: db) ?? false
            if conditionResolved {
                return resolve(expression: trueExp, context: context, db: db)
            } else {
                return resolve(expression: falseExp, context: context, db: db)
            }
        case let .or(a, b):
            return resolve(expression: a, context: context, db: db) ?? resolve(expression: b, context: context, db: db)
        case .negation(_):
            print("CVU Expression error: Should not use ! operator on non-boolean value")
            return nil
        case let .addition(a, b):
            return (resolve(expression: a, context: context, db: db) ?? 0) + (resolve(expression: b, context: context, db: db) ?? 0)
        case let .subtraction(a, b):
            return (resolve(expression: a, context: context, db: db) ?? 0) - (resolve(expression: b, context: context, db: db) ?? 0)
        case let .constant(const):
            return const.asNumber()
        case let .multiplication(a, b):
            return (resolve(expression: a, context: context, db: db) ?? 0) * (resolve(expression: b, context: context, db: db) ?? 0)
        case let .division(a, b):
            guard let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
                  let rhs = resolve(expression: b, context: context, type: Double.self, db: db),
                  rhs != 0
            else { return nil }
            return lhs / rhs
        default:
//            print("CVU Expression: \(expression) unsupported for resolving to double")
            return nil
        }
    }
    
    func resolve(expression: CVUValue.ExpressionNode, context: CVUContext, type: String.Type = String.self, db: DatabaseController) -> String? {
        switch expression {
        case let .lookup(nodes):
            return resolve(nodes: nodes, context: context, db: db)
        case let .conditional(condition, trueExp, falseExp):
            let conditionResolved = resolve(expression: condition, context: context, type: Bool.self, db: db) ?? false
            if conditionResolved {
                return resolve(expression: trueExp, context: context, db: db)
            } else {
                return resolve(expression: falseExp, context: context, db: db)
            }
        case let .or(a, b):
            return resolve(expression: a, context: context, type: String.self, db: db)?.nilIfBlank ?? resolve(expression: b, context: context, db: db)
        case .negation(_):
            print("CVU Expression error: Should not use ! operator on non-boolean value")
            return nil
        case let .addition(a, b):
            return (resolve(expression: a, context: context, db: db) ?? "") + (resolve(expression: b, context: context, db: db) ?? "")
        case .subtraction(_, _):
            print("CVU Expression error: Should not use - operator on string value")
            return nil
        case let .constant(const):
            return const.asString()
        case let .stringMode(nodes: nodes):
            return nodes.compactMap { resolve(expression: $0, context: context, db: db) }.joined(separator: "")
        default:
//            print("CVU Expression: \(expression) unsupported for resolving to string")
            return nil
        }
    }
    
    func resolve(expression: CVUValue.ExpressionNode, context: CVUContext, type: Bool.Type = Bool.self, db: DatabaseController) -> Bool? {
        switch expression {
        case let .lookup(nodes):
            return resolve(nodes: nodes, context: context, db: db)
        case let .conditional(condition, trueExp, falseExp):
            let conditionResolved = resolve(expression: condition, context: context, type: Bool.self, db: db) ?? false
            if conditionResolved {
                return resolve(expression: trueExp, context: context, db: db)
            } else {
                return resolve(expression: falseExp, context: context, db: db)
            }
        case let .and(a, b):
            return (resolve(expression: a, context: context, db: db) ?? false) && (resolve(expression: b, context: context, db: db) ?? false)
        case let .or(a, b):
            return (resolve(expression: a, context: context, db: db) ?? false) || (resolve(expression: b, context: context, db: db) ?? false)
        case let .negation(x):
            return resolve(expression: x, context: context, type: Bool.self, db: db).map { !$0 }
        case .addition(_ ,_):
            print("CVU Expression error: Should not use + operator on bool value")
            return nil
        case .subtraction(_, _):
            print("CVU Expression error: Should not use - operator on bool value")
            return nil
        case let .constant(const):
            return const.asBool()
        case let .lessThan(a, b):
            guard let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
                  let rhs = resolve(expression: b, context: context, type: Double.self, db: db)
            else { return nil }
            return lhs < rhs
        case let .greaterThan(a, b):
            guard let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
                  let rhs = resolve(expression: b, context: context, type: Double.self, db: db)
            else { return nil }
            return lhs > rhs
        case let .lessThanOrEqual(a, b):
            guard let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
                  let rhs = resolve(expression: b, context: context, type: Double.self, db: db)
            else { return nil }
            return lhs <= rhs
        case let .greaterThanOrEqual(a, b):
            guard let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
                  let rhs = resolve(expression: b, context: context, type: Double.self, db: db)
            else { return nil }
            return lhs >= rhs
        case let .areEqual(a, b):
            if let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
               let rhs = resolve(expression: b, context: context, type: Double.self, db: db) {
                return lhs == rhs
            } else if let lhs = resolve(expression: a, context: context, type: String.self, db: db),
                      let rhs = resolve(expression: b, context: context, type: String.self, db: db) {
                return lhs == rhs
            } else if let lhs = resolve(expression: a, context: context, type: Bool.self, db: db),
                      let rhs = resolve(expression: b, context: context, type: Bool.self, db: db) {
                return lhs == rhs
            } else { return false }
        case let .areNotEqual(a, b):
            if let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
               let rhs = resolve(expression: b, context: context, type: Double.self, db: db) {
                return lhs != rhs
            } else if let lhs = resolve(expression: a, context: context, type: String.self, db: db),
                      let rhs = resolve(expression: b, context: context, type: String.self, db: db) {
                return lhs != rhs
            } else if let lhs = resolve(expression: a, context: context, type: Bool.self, db: db),
                      let rhs = resolve(expression: b, context: context, type: Bool.self, db: db) {
                return lhs != rhs
            } else { return true }
        default:
//            print("CVU Expression: \(expression) unsupported for resolving to bool")
            return nil
        }
    }
}

import SwiftUI
extension CVULookupController {
    func resolveToItemAndProperty(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> (item: ItemRecord, property: String)? {
        var currentItem: ItemRecord?
        
        // Find the item referenced by the lookup
        for node in nodes.dropLast() {
            switch node.type {
            case .default:
                guard let defaultItem = context.currentItem else { return nil }
                currentItem = defaultItem
            case .lookup(let subexpression):
                if let nextItem = currentItem {
                    let step = itemLookup(node: node, items: [nextItem], subexpression: subexpression, context: context, db: db)
                    if case let .items(items) = step {
                        currentItem = items.first
                    }
                }
            case .function:
            return nil
            }
        }
        
        guard let targetItem = currentItem, let propertyLookup = nodes.last else { return nil }
        // Make a binding to the right property
        return (item: targetItem, property: propertyLookup.name)
    }
}
