//
//  PodAPIRequests.swift
//  MemriDatabase
//
//  Created by T Brennan on 18/12/20.
//

import Foundation
import Alamofire
import Combine

struct RequestBody<Payload: Encodable>: Encodable {
    init(connectionConfig: PodAPIConnectionDetails, payload: Payload) {
        self.databaseKey = connectionConfig.databaseKey
        self.payload = payload
    }
    
    var databaseKey: String
    var payload: Payload
    
    
}


struct PodAPIStandardRequest<Payload: Encodable> {
    var method: HTTPMethod = .post
    var path: String
    var headers: HTTPHeaders = [:]
    var payload: Payload
    
    func constructRequest(connectionConfig: PodAPIConnectionDetails) throws -> URLRequest {
        var components = URLComponents()
        components.scheme = connectionConfig.scheme
        components.host = connectionConfig.host
        components.port = connectionConfig.port
        components.path = "/\(connectionConfig.apiVersion)/\(connectionConfig.ownerKey)/\(path)"
        
        guard let url = components.url else {
            throw StringError(description: "Failed to construct URL for APIRequest")
        }
        
        print(url)
        
        var request = URLRequest(url: url)
        request.method = method
        
        switch method {
        case .get:
            /// For a `get` request, encode our payload into the URL.
            if !(Payload.self == PodAPIPayload.EmptyPayload.self) {
                request = try URLEncodedFormParameterEncoder().encode(payload, into: request)
            }
        default:
            /// For a `post` request (or other types of request) encode our payload into the body of the request.
            let body = RequestBody(connectionConfig: connectionConfig,
                                   payload: payload)
            request = try JSONParameterEncoder().encode(body, into: request)
        }
        
        return request
    }
    
    private func executeRequest(connectionConfig: PodAPIConnectionDetails) throws -> DataRequest {
        let request = AF.request(try constructRequest(connectionConfig: connectionConfig))
            .validate(statusCode: 200..<300)
        return request
    }
    
    func execute(connectionConfig: PodAPIConnectionDetails) throws -> AnyPublisher<DataResponse<Any, AFError>, Never> {
        try executeRequest(connectionConfig: connectionConfig)
            .publishResponse(using:  JSONResponseSerializer()).eraseToAnyPublisher()
    }

    func execute<T: Decodable>(connectionConfig: PodAPIConnectionDetails, decodeAs: T.Type = T.self) throws -> AnyPublisher<DataResponse<T, AFError>, Never> {
        try executeRequest(connectionConfig: connectionConfig)
            .publishDecodable().eraseToAnyPublisher()
    }
}

struct PodAPIUploadRequest<Payload: Encodable> {
    var path: String
    var fileURL: URL
    var payload: Payload
    
    func constructRequest(connectionConfig: PodAPIConnectionDetails) throws -> URLRequest {
        var components = URLComponents()
        components.scheme = connectionConfig.scheme
        components.host = connectionConfig.host
        components.port = connectionConfig.port
        components.path = "/\(connectionConfig.apiVersion)/\(connectionConfig.ownerKey)/\(path)"
        
        guard let url = components.url else {
            throw StringError(description: "Failed to construct URL for APIRequest")
        }
        
        var request = URLRequest(url: url)
        request.method = .post
        if !(Payload.self == PodAPIPayload.EmptyPayload.self) {
            request = try URLEncodedFormParameterEncoder().encode(payload, into: request)
        }
        
        return request
    }
    
    func execute(connectionConfig: PodAPIConnectionDetails) throws -> AnyPublisher<DataResponse<Any, AFError>, Never> {
        AF.upload(fileURL, with: try constructRequest(connectionConfig: connectionConfig))
            .publishResponse(using:  JSONResponseSerializer()).eraseToAnyPublisher()
    }
    
    func execute<T: Decodable>(connectionConfig: PodAPIConnectionDetails) throws -> AnyPublisher<DataResponse<T, AFError>, Never> {
        AF.upload(fileURL, with: try constructRequest(connectionConfig: connectionConfig))
            .publishDecodable().eraseToAnyPublisher()
    }
}

struct PodAPIDownloadRequest<Payload: Encodable> {
    var path: String
    var method: HTTPMethod = .post
    var payload: Payload
    var fileUID: String
    
    var destination: DownloadRequest.Destination { { _, _ in
        let fileURL = FileStorageController.getURLForFile(withUUID: fileUID)
        return (fileURL, [.removePreviousFile, .createIntermediateDirectories])
    }
    }
    
    func constructRequest(connectionConfig: PodAPIConnectionDetails) throws -> URLRequest {
        var components = URLComponents()
        components.scheme = connectionConfig.scheme
        components.host = connectionConfig.host
        components.port = connectionConfig.port
        components.path = "/\(connectionConfig.apiVersion)/\(connectionConfig.ownerKey)/\(path)"
        
        guard let url = components.url else {
            throw StringError(description: "Failed to construct URL for APIRequest")
        }
        
        print(url)
        
        var request = URLRequest(url: url)
        request.method = method
        
        switch method {
        case .get:
            /// For a `get` request, encode our payload into the URL.
            if !(Payload.self == PodAPIPayload.EmptyPayload.self) {
                request = try URLEncodedFormParameterEncoder().encode(payload, into: request)
            }
        default:
            /// For a `post` request (or other types of request) encode our payload into the body of the request.
            let body = RequestBody(connectionConfig: connectionConfig,
                                   payload: payload)
            request = try JSONParameterEncoder().encode(body, into: request)
        }
        
        return request
    }
    
    func execute(connectionConfig: PodAPIConnectionDetails) throws -> AnyPublisher<DownloadResponse<URL, AFError>, Never> {
        AF.download(try constructRequest(connectionConfig: connectionConfig), to: destination)
            .publishURL().eraseToAnyPublisher()
    }
}

extension PodAPIStandardRequest {
    static func runImporter(importerUID: StringUUID, connectionConfig: PodAPIConnectionDetails) -> PodAPIStandardRequest where Payload == PodAPIPayload.ItemUIDWithServicePayload {
        let payload = PodAPIPayload.ItemUIDWithServicePayload(uid: importerUID, servicePayload: .init(connectionConfig: connectionConfig))
        return PodAPIStandardRequest(path: "run_importer", payload: payload)
    }
    
    static func runIndexer(indexerUID: StringUUID, connectionConfig: PodAPIConnectionDetails) -> PodAPIStandardRequest where Payload == PodAPIPayload.ItemUIDWithServicePayload {
        let payload = PodAPIPayload.ItemUIDWithServicePayload(uid: indexerUID, servicePayload: .init(connectionConfig: connectionConfig))
        return PodAPIStandardRequest(path: "run_importer", payload: payload)
    }
    
    static func getItemWithEdges(itemUID: Payload) -> PodAPIStandardRequest {
        // Note payload is just item UID (no JSON)
        let payload = itemUID
        return PodAPIStandardRequest(path: "get_item", payload: payload)
    }
    
    static func getItemsWithEdges(itemUIDs: Set<StringUUID>) -> PodAPIStandardRequest where Payload == PodAPIPayload.ItemUIDList {
        let payload = PodAPIPayload.ItemUIDList(uids: itemUIDs)
        return PodAPIStandardRequest(path: "get_items_with_edges", payload: payload)
    }
    
    static func searchByFields(itemType: String) -> PodAPIStandardRequest where Payload == [String: String] {
        let payload = ["_type": itemType]
        return PodAPIStandardRequest(path: "search_by_fields", payload: payload)
    }
    
    static func getAllItems() -> PodAPIStandardRequest where Payload == PodAPIPayload.EmptyPayload {
        return PodAPIStandardRequest(path: "get_all_items", payload: PodAPIPayload.EmptyPayload())
    }
    
    static func createItem(syncDict: [String: AnyEncodable]) -> PodAPIStandardRequest where Payload == [String: AnyEncodable] {
        return PodAPIStandardRequest(path: "create_item", payload: syncDict)
    }
    
    static func deleteItem(itemUID: StringUUID) -> PodAPIStandardRequest where Payload == StringUUID {
        // Note payload is just item UID (no JSON)
        let payload = itemUID
        return PodAPIStandardRequest(path: "delete_item", payload: payload)
    }
    
    static func bulkAction(payload: Payload) -> PodAPIStandardRequest where Payload == PodAPIPayload.BulkAction {
        return PodAPIStandardRequest(path: "bulk_action", payload: payload)
    }
}

extension PodAPIDownloadRequest {
    static func downloadFile(fileSHAHash: String, fileUID: String) -> PodAPIDownloadRequest where Payload == PodAPIPayload.FileSHA {
        PodAPIDownloadRequest(path: "get_file", payload: PodAPIPayload.FileSHA(sha: fileSHAHash), fileUID: fileUID)
    }
}

extension PodAPIUploadRequest {
    static func uploadFile(fileURL: URL, fileSHAHash: String? = nil, connectionConfig: PodAPIConnectionDetails) throws -> PodAPIUploadRequest where Payload == PodAPIPayload.EmptyPayload {
        let sha = try fileSHAHash ?? FileHelper.getSHAHash(url: fileURL)
        let path = "upload_file/\(connectionConfig.databaseKey)/\(sha)"
        
        return PodAPIUploadRequest(path: path, fileURL: fileURL, payload: PodAPIPayload.EmptyPayload())
    }
}

/*
 
 /// SEE POD API DOCUMENTATION AT https://gitlab.memri.io/memri/pod/-/blob/dev/docs/HTTP_API.md
 
 
 
 static func searchByFields() -> APIRequest {
     APIRequest(path: "search_by_fields")
 }
 
 /// NOTE: This API is atomic. If any part of it fails, no changes will be applied
 static func bulkAction() -> APIRequest {
     return APIRequest(path: "bulk_action")
 }
 
 static func uploadFile(fileSHA256: String, localFileURL: URL) -> APIRequest {
     APIRequest(path: "upload_file/\(fileSHA256)",
                payload: .fileURL(localFileURL))
 }
 
 static func downloadFile(get_file: String) -> APIRequest {
     APIRequest(path: "get_file", payload: .json(APITypes.FileSHAPayload(sha: get_file)))
 }
 */
