//
//  AppController.swift
//  MemriDatabase
//
//  Created by T Brennan on 14/12/20.
//

import Foundation
import KeychainAccess

class AppController: ObservableObject {
    static let shared = AppController()
    
    let databaseController: DatabaseController
    let syncController: SyncController
    let cvuController: CVUController

    private init() {
        let databaseController = DatabaseController()
        self.databaseController = databaseController
        self.syncController = SyncController(databaseController: databaseController)
        self.cvuController = CVUController()
    }
    
    func onLaunch() {
        #if DEBUG
//        setHasBeenSetup(databaseKey: nil)
        #endif
        
        requestAuthentication()
    }
    
    enum AppState {
        case setup
        case authentication
        case authenticated
    }
    
    @Published
    var state: AppState = .setup
    
    private func updateState() {
        guard checkHasBeenSetup() else { state = .setup; return }
        guard isAuthenticated else {
            state = .authentication;
            requestAuthentication()
            return
        }
        
        state = .authenticated
    }
    
    // MARK: Setup
    func setupApp(config: SetupConfig, onCompletion: @escaping (_ error: Error?) -> Void) {
        DispatchQueue.global(qos: .background).async {
            switch config {
            case .local, .newPod:
                do {
                    guard !self.databaseController.databaseIsSetup else {
                        // If there is already data set up, don't import
                        onCompletion(nil)
                        return
                    }
                    try DemoData.importDemoData(databaseController: self.databaseController)
                }
                catch {
                    onCompletion(error)
                    return
                }
            case .existingPod:
                break
            }
            /// During this setup function would be a good place to generate a database encryption key, create a new database with this key, and then import the demo data.
            /// NOTE: This is a temporary placehold until encryption is implemented.
            /// - UUID is not a good option for a randomly generated key, should use an existing generator from CryptoKit
            let newDatabaseEncryptionKey = UUID().uuidString
            self.setHasBeenSetup(databaseKey: newDatabaseEncryptionKey)
            onCompletion(nil)
        }
    }
    
    
    // MARK: Authentication
    private(set) var isAuthenticated: Bool = false {
        didSet { if isAuthenticated != oldValue { updateState() }}
    }
    
    func requestAuthentication() {
        guard checkHasBeenSetup() else { return }
        if let dbKey = try? Keychain().getString(AppController.keychain_databaseKey) {
            print("GOT KEY: \(dbKey)")
            isAuthenticated = true
        } else {
            isAuthenticated = false
            print("NEEDS AUTH")
        }
    }
    
    static let keychain_databaseKey: String = "memri_databaseKey"
    func checkHasBeenSetup() -> Bool {
        guard (try? Keychain().contains(AppController.keychain_databaseKey, withoutAuthenticationUI: true)) ?? false else {
            return false
        }
        guard AppController.shared.databaseController.databaseIsSetup else { return false }
        return true
    }
    
    func setHasBeenSetup(databaseKey: String?) {
        if let databaseKey = databaseKey {
            try? Keychain().set(databaseKey, key: AppController.keychain_databaseKey)
        } else {
            try? Keychain().remove(AppController.keychain_databaseKey)
        }
        DispatchQueue.main.async {
            self.updateState()
        }
    }
    
    // MARK: Pod connection
    var podConnectionConfig: PodAPIConnectionDetails? {
        do {
            return try databaseController.read { db in
                // Here you should retrieve the connection details stored in the database
                return PodAPIConnectionDetails()
            }
        } catch {
            print(error)
            return nil
        }
    }
}


enum SetupConfig {
    case local
    case newPod(NewPodConfig)
    case existingPod(ExistingPodConfig)
    
    struct NewPodConfig {
        var podURL: String
    }
    
    struct ExistingPodConfig {
        var podURL: String
        var podPrivateKey: String
        var podPublicKey: String
        var podDatabaseKey: String
    }
}
