//
//  GRDBTest.swift
//  MemriDatabase
//
//  Created by T Brennan on 10/11/20.
//

import Foundation
import GRDB

/// The database controller provides access to the app's SQLite database. Generally only a single database controller will be used throughout the app
class DatabaseController {
    /// Create a DatabaseController. Change the databaseName to create/access a different database file (eg. for testing purposes)
    init(databaseName: String = "memri", schema: Schema = try! .loadFromFile()) {
        self.schema = schema
        databasePool = {
            let paths = FileManager.default.urls(for: .documentDirectory, in: .userDomainMask)
            let documentsDirectory = paths[0]
            let url = documentsDirectory.appendingPathComponent("\(databaseName).sqlite")
            print(url)
            return try! DatabasePool(path: url.path)
        }()
        
        if !Constants.isRunningTests {
            try! setupDatabase()
        }
    }
    
    private(set) var schema: Schema
    
    /// This variable is `thread-local` and used to track whether there is a database `read` already in progress
    private var readSession: ThreadLocal<Database?> = .init(value: nil)
    
    /// This variable is `thread-local` and used to track whether there is a database `write` already in progress
    private var writeSession: ThreadLocal<Database?> = .init(value: nil)
    
    /// This is the connection to the database used throughout the app
    /// It uses SQLite's WAL mode to provide concurrent reads and allow for read/write from background threads
    let databasePool: DatabasePool
    
    /// Check if the database has been setup
    var databaseIsSetup: Bool {
        (try? read { db in
            try ItemRecord.fetchOne(db) != nil
        }) ?? false
    }
    
    /// Prepare the database and register migration steps.
    func setupDatabase() throws {
        var migrator = DatabaseMigrator()
        
        if Constants.isRunningTests {
            /// *NEVER* set this outside of TESTING config (because it deletes all data)
            try databasePool.erase()
        }
        
        #if DEBUG
        /// Aid development by wiping the database EVERY LAUNCH
        /// *NEVER* set this outside of DEBUG config (because it deletes all data)
//        try databasePool.erase()
        
        /// Aid database development by wiping the database when migrations change
        /// *NEVER* set this outside of DEBUG config (because it deletes all data)
        migrator.eraseDatabaseOnSchemaChange = true
        #endif
        
        migrator.registerMigration("initialSetup") { db in
            try db.create(table: "item") { t in
                /// 16 byte UUID that uniquely identifies this item
                t.column("uid", .text).notNull().primaryKey()
                
                /// The type of this item (a string matching a type defined in the schema)
                t.column("type", .text).notNull()
                
                /// Time of creation
                t.column("dateCreated", .integer).notNull()
                
                /// Time last modified
                t.column("dateModified", .integer).notNull()
                
                /// Version - starts at 1
                t.column("version", .integer).defaults(to: 1).notNull(onConflict: .replace)
                
                /// Whether this item has been deleted (default = false)
                t.column("deleted", .boolean).defaults(to: false).notNull(onConflict: .replace)
                
                /// iOS only (used for syncing decisions
                t.column("syncState", .text).defaults(to: SyncState.create).notNull(onConflict: .replace)
                t.column("syncHasPriority", .boolean).defaults(to: false).notNull(onConflict: .replace)
            }
            
            try db.create(table: "itemEdge") { t in
                /// The uid of the item this edge belongs to. Delete this row if the item is deleted
                t.column("sourceUID", .text).notNull().indexed().references("item", onDelete: .cascade)
                
                /// The name of this edge
                t.column("name", .text).notNull()
                
                /// The target item of this edge. Delete this row if the target item is deleted
                t.column("targetUID", .text).notNull().indexed().references("item", onDelete: .cascade)
            }
            
            try db.create(table: "itemProperty") { t in
                /// The uid of the item this property belongs to. Delete this row if the item is deleted
                t.column("itemUID", .text).notNull().indexed().references("item", onDelete: .cascade)
                
                /// The name of this property
                t.column("name", .text).notNull()
                
                /// The value of this property
                t.column("value", .blob).notNull()
                
                
                
                /// There must be only one row for each property name per item - if we try to add another replace the old one
                t.uniqueKey(["itemUID", "name"], onConflict: .replace)
            }
            
            try db.create(virtualTable: "itemProperty_search", using: FTS4()) { t in
                t.synchronize(withTable: "itemProperty")
                t.tokenizer = .porter
                t.column("itemUID").notIndexed()
                t.column("name").notIndexed()
                t.column("value")
            }
            
            try db.create(table: "navigationState") { t in
                t.column("sessionID", .text).notNull().primaryKey()
                t.column("state", .blob).notNull()
            }
        }
        
        
        
        try migrator.migrate(databasePool)
    }
    
    
    func search(string searchString: String) throws -> [ItemRecord] {
        try read { (db) in
            guard let searchQuery = FTS3Pattern(matchingAllTokensIn: searchString) else {
                throw StringError(description: "Invalid search string: \(searchString)")
            }
            let refinedQuery = try FTS3Pattern(rawPattern: "\(searchQuery.rawPattern)*")
            let search = try ItemRecord.search(db, pattern: refinedQuery)
            return search
        }
    }
}

import SwiftUI
extension DatabaseController {
    /// Read from the database while constructing a SwiftUI view
    func read<Content: View>(@ViewBuilder content: (Database) throws -> Content) -> some View {
        if let existingReadDB = readSession.inner.value {
            return try? content(existingReadDB)
        } else {
            defer {
                readSession.inner.value = nil
            }
            return try? databasePool.read { (db) -> Content? in
                readSession.inner.value = db
                return try? content(db)
            }
        }
    }
    
    /// Read from the database and return a value
    func read<T>(_ block: (Database) throws -> T ) throws -> T {
        if let existingReadDB = readSession.inner.value {
            return try block(existingReadDB)
        } else {
            defer {
                readSession.inner.value = nil
            }
            return try databasePool.read { (db) -> T in
                readSession.inner.value = db
                return try block(db)
            }
        }
    }
    
    // Write to the database. This happens async by default
    func write(_ updates: @escaping (Database) throws -> Void ) {
        DispatchQueue.global(qos: .userInteractive).async {
            do {
                try self.databasePool.write { (db) in
                    try updates(db)
                }
            } catch {
                print(error)
            }
        }
    }
    
    // Write to the database immediately. Note that this may throw an error if inside a database read
    func writeSync<T>(_ updates: (Database) throws -> T ) throws -> T {
        if let existingWriteDB = writeSession.inner.value {
            return try updates(existingWriteDB)
        } else {
            defer {
                writeSession.inner.value = nil
            }
            
            return try databasePool.write { (db) in
                writeSession.inner.value = db
                return try updates(db)
            }
        }
    }
}
