//
//  DatabaseQuery.swift
//  MemriDatabase
//
//  Created by T Brennan on 7/12/20.
//

import Foundation
import GRDB
import Combine


/// This type is used to describe a database query. It can be stored as a JSON (it is codable)
struct DatabaseQueryConfig: Equatable, Codable {
    /// A list of item types to include. Default is Empty -> ALL item types
    var itemTypes: [String] = ["Person", "Note", "Address", "Photo", "Indexer", "Importer"]
    
    /// A list of item UIDs to include. Default is Empty -> don't filter on UID
    var itemUIDs: Set<StringUUID> = []
    
    /// A property to sort the results by
    var sortProperty: String? = "dateModified"
    
    /// Only include items modified after this date
    var dateModifiedAfter: Date?
    
    /// Only include items modified before this date
    var dateModifiedBefore: Date?
    
    /// Only include items created after this date
    var dateCreatedAfter: Date?
    
    /// Only include items created before this date
    var dateCreatedBefore: Date?
    
    /// The maximum number of items to fetch
    var pageSize: Int = 1000
    
    /// Use this to change which page is requested by the query (eg. if there are more than `pageSize` items)
    var currentPage: Int = 0
    
    /// A search string to match item properties against
    var searchString: String? = nil
    
    /// If enabled the search will find items that link to another item matching the search term. This only goes one edge deep for performance purposes.
    var includeImmediateEdgeSearch: Bool = true
    
    /// A list of conditions. eg. property name = "Demo note"
    var conditions: [DatabaseQueryCondition] = []
    
    private func constructFilteredRequest(withSearchUIDs searchUIDs: Set<StringUUID>?) -> QueryInterfaceRequest<ItemRecord> {
        let limit = pageSize
        let offset = pageSize * currentPage
        
        var queryConditions: [SQLExpression] = []
        /// Filter by item type
        if !itemTypes.isEmpty {
            queryConditions.append(itemTypes.contains(ItemRecord.Columns.type))
        }
        
        
        /// Filter to only include items matching the search term (AND if already filtered by UID, those that match both)
        if let searchUIDs = searchUIDs {
            if !itemUIDs.isEmpty {
                let combinedUIDs = searchUIDs.intersection(itemUIDs)
                queryConditions.append(combinedUIDs.contains(ItemRecord.Columns.uid))
            } else {
                queryConditions.append(searchUIDs.contains(ItemRecord.Columns.uid))
            }
        } else if !itemUIDs.isEmpty {
            queryConditions.append(itemUIDs.contains(ItemRecord.Columns.uid))
        }
        
        /// Filter by date ranges
        if let dateModifiedBefore = dateModifiedBefore {
            queryConditions.append(ItemRecord.Columns.dateModified <= dateModifiedBefore)
        }
        if let dateModifiedAfter = dateModifiedAfter {
            queryConditions.append(ItemRecord.Columns.dateModified >= dateModifiedAfter)
        }
        if let dateCreatedBefore = dateCreatedBefore {
            queryConditions.append(ItemRecord.Columns.dateModified <= dateCreatedBefore)
        }
        if let dateCreatedAfter = dateCreatedAfter {
            queryConditions.append(ItemRecord.Columns.dateModified >= dateCreatedAfter)
        }
        
        var request = ItemRecord.limit(limit, offset: offset).filter(queryConditions.joined(operator: .and)).distinct()
        
        // Property conditions
        let queryPropertyConditions: [SQLExpression] = conditions.compactMap { condition in
            switch condition {
            case let .propertyEquals(info):
                return ItemPropertyRecord.Columns.name == info.name && ItemPropertyRecord.Columns.value == info.value
            default:
                return nil
            }
        }
        if !queryPropertyConditions.isEmpty {
            let propertyFilter = ItemRecord.properties.filter(queryPropertyConditions.joined(operator: .and))
            request = request.having(propertyFilter.isEmpty == false)
        }
        
        // Edge conditions
        let queryEdgeConditions: [SQLExpression] = conditions.compactMap { condition in
            switch condition {
            case let .edgeHasTarget(info):
                return ItemEdgeRecord.Columns.name == info.edgeName && ItemEdgeRecord.Columns.targetUID == info.targetUID
            default:
                return nil
            }
        }
        if !queryEdgeConditions.isEmpty {
            let edgeFilter = ItemRecord.edges.filter(queryEdgeConditions.joined(operator: .and))
            request = request.having(edgeFilter.isEmpty == false)
        }
        
        switch sortProperty?.nilIfBlank {
        case "dateCreated":
            request = request.order([ItemRecord.Columns.dateCreated, ItemRecord.Columns.dateModified])
        case "dateModified":
            request = request.order([ItemRecord.Columns.dateModified, ItemRecord.Columns.dateCreated])
        case .some(let sortProperty):
            let sortAlias = TableAlias()
            request = request.joining(optional:
                                        ItemRecord.properties.aliased(sortAlias)
                                        .filter(ItemPropertyRecord.Columns.name == sortProperty)
            )
            .order([sortAlias[ItemPropertyRecord.Columns.value].ascNullsLast,
                    ItemRecord.Columns.dateModified,
                    ItemRecord.Columns.dateCreated])
        default: break
        }
        
        return request
    }
    
    private func constructSearchRequest() -> SQLRequest<StringUUID?>? {
        guard let searchString = searchString?.nilIfBlank,
              let searchQuery = FTS3Pattern(matchingAllTokensIn: searchString),
              let refinedQuery = try? FTS3Pattern(rawPattern: "\(searchQuery.rawPattern)*")
        else {
            return nil
        }
        return SQLRequest(sql:
            """
               SELECT DISTINCT \(ItemPropertyRecord.Columns.itemUID)
               FROM \(ItemPropertyRecord.databaseSearchTableName)
               WHERE \(ItemPropertyRecord.databaseSearchTableName) MATCH ?
            """, arguments: [refinedQuery])
    }
    
    func executeRequest(db: Database) throws -> [ItemRecord] {
        var searchUIDs = try constructSearchRequest().map { try Set($0.fetchAll(db).compactMap { $0 }) }
        if includeImmediateEdgeSearch {
            /// Find items connected the the search results by one or more edges. Eg. if a file is found based on the search term, we will also include a Photo or Note that links to it
            let edgeUIDs = try searchUIDs.map { searchUIDs -> [StringUUID] in
                try ItemEdgeRecord.filter(searchUIDs.contains(ItemEdgeRecord.Columns.targetUID)).select([ItemEdgeRecord.Columns.sourceUID]).fetchAll(db)
            }
            searchUIDs = searchUIDs?.union(edgeUIDs ?? [])
        }
        
        let filterRequest = constructFilteredRequest(withSearchUIDs: searchUIDs)
        
        return try filterRequest.fetchAll(db)
    }
}

enum DatabaseQueryCondition: Codable, Equatable {
    case propertyEquals(PropertyEquals)
    case edgeHasTarget(EdgeHasTarget)
    
    enum CodingKeys: CodingKey {
        case propertyEquals, edgeHasTarget
    }
    
    struct PropertyEquals: Codable, Equatable {
        var name: String
        var value: String
    }
    
    struct EdgeHasTarget: Codable, Equatable {
        var edgeName: String
        var targetUID: StringUUID
    }
    
    func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        switch self {
        case .propertyEquals(let info):
            try container.encode(info, forKey: .propertyEquals)
        case .edgeHasTarget(let info):
            try container.encode(info, forKey: .edgeHasTarget)
        }
    }
    
    init(from decoder: Decoder) throws {
        let container = try decoder.container(keyedBy: CodingKeys.self)
        let key = container.allKeys.first
        
        switch key {
        case .propertyEquals:
            try self = .propertyEquals(container.decode(PropertyEquals.self, forKey: .propertyEquals))
        case .edgeHasTarget:
            try self = .edgeHasTarget(container.decode(EdgeHasTarget.self, forKey: .edgeHasTarget))
        case .none:
            throw StringError(description: "Unknown database query condition")
        }
    }
}
