//
//  DemoData.swift
//  MemriDatabase
//
//  Created by T Brennan on 30/11/20.
//

import Foundation

enum DemoData {
    static func importDemoData(databaseController: DatabaseController = AppController.shared.databaseController, throwIfAgainstSchema: Bool = false) throws {
        let fileURL = Bundle.main.url(forResource: "demo_database", withExtension: "json")!
        let fileData = try Data(contentsOf: fileURL)
        guard let items = try JSONSerialization.jsonObject(with: fileData, options: []) as? [[String: Any]] else {
            return
        }
        
        let processedItems: [DemoData_Item] = try items.flatMap { try processItemJSON(item: $0, schema: databaseController.schema) }
        
        try databaseController.writeSync { (db) in
            var tempUIDLookup: [String: StringUUID] = [:]
            
            for item in processedItems {
                let record = ItemRecord(uid: item.UID, type: item.type, dateCreated: item.dateCreated, dateModified: item.dateModified)
                if let tempUID = item.tempUID {
                    tempUIDLookup[tempUID] = record.uid
                }
                try record.insert(db)
            }
            
            for item in processedItems {
                for property in item.properties {
                    let record = ItemPropertyRecord(itemUID: item.UID, name: property.name, value: property.value)
                    try record.insert(db)
                }
                for edge in item.edges {
                    guard let targetActualUID = tempUIDLookup[edge.targetTempUID] else { continue }
                    let record = ItemEdgeRecord(sourceUID: item.UID, name: edge.name, targetUID:targetActualUID)
                    try record.insert(db)
                }
            }
        }
    }


    static func processItemJSON(item: [String: Any], schema: Schema, overrideUID: String? = nil) throws -> [DemoData_Item] {
        func handleError(_ string: String) throws {
            if Constants.isRunningTests {
                // Used for testing: throw an error if error in demo data
                throw(StringError(description: string))
            } else {
                // Notify developer of error in demo data, but continue
                print(string)
            }
        }
        
        guard let itemType = item["_type"] as? String else {
            try handleError("BAD RECORD: \(item)")
            return []
        }
        guard schema.types[itemType] != nil else {
            print("\(itemType) not in schema")
            return []
        }
        
        let itemTempUID = overrideUID ?? (item["uid"] as? Int).map(String.init)
        
        var items: [DemoData_Item] = []
        var properties: [DemoData_Property] = []
        var edges: [DemoData_Edge] = []
        
        // Fake a recent date for the demo data
        var dateCreated = Date(timeIntervalSinceNow: -TimeInterval.random(in: 0...1814400))
        var dateModified = dateCreated
        
        propertyIterator: for (propertyName, propertyValue) in item {
            switch propertyName {
            case "_type", "uid", "version":
                continue
            case "allEdges":
                guard let edgeList = propertyValue as? [[String: Any]] else { continue }
                for edge in edgeList {
                    guard let edgeName = edge["_type"] as? String else { continue }
                    if let targetUID = (edge["uid"] as? Int).map(String.init) {
                        edges.append(DemoData_Edge(name: edgeName, targetTempUID: targetUID))
                    } else if let subitem = edge["_target"] as? [String: Any] {
                        // Sub-item declared as edge, add edge property AND item
                        let targetUID = UUID().uuidString
                        edges.append(DemoData_Edge(name: edgeName, targetTempUID: targetUID))
                        items.append(contentsOf: try processItemJSON(item: subitem, schema: schema, overrideUID: targetUID))
                    }
                }
            case "dateCreated":
                guard let dateInMilliseconds = propertyValue as? Int, let date = DatabaseHelperFunctions.decode(from: dateInMilliseconds)
                else {
                    continue
                }
                dateCreated = date
                dateModified = date
            default:
                guard let expectedType = schema.types[itemType]?.propertyTypes[propertyName] else {
                    try handleError("\(itemType).\(propertyName) property not in schema")
                    continue
                }
                do {
                    let result = try PropertyDatabaseValue(value: propertyValue, propertyType: expectedType.valueType, debugInfo: "\(itemType).\(propertyName)")
                    properties.append(DemoData_Property(name: propertyName, value: result))
                }
                catch {
                    try handleError(error.localizedDescription)
                    continue
                }
            }
        }
        
        items.append(DemoData_Item(type: itemType, tempUID: itemTempUID , properties: properties, edges: edges, dateCreated: dateCreated, dateModified: dateModified))
        return items
    }

    struct DemoData_Item {
        var type: String
        var UID: StringUUID = StringUUID()
        var tempUID: String?
        var properties: [DemoData_Property]
        var edges: [DemoData_Edge]
        var dateCreated: Date?
        var dateModified: Date?
    }

    struct DemoData_Property {
        var name: String
        var value: PropertyDatabaseValue
    }

    struct DemoData_Edge {
        var name: String
        var targetTempUID: String
    }

}
