//
//  DatabaseTypes.swift
//  MemriDatabase
//
//  Created by T Brennan on 13/11/20.
//

import Foundation
import GRDB
import CoreGraphics


/// This type is used as an intermediate for storing and retrieving values from the database
/// It makes use of the dynamic schema to determine what type to resolve the data as.
enum PropertyDatabaseValue: Equatable {
    case string(String)
    case bool(Bool)
    case int(Int)
    case double(Double)
    case datetime(Date)
    case blob(Data)
    
    init?(databaseValue: DatabaseValue, propertyType: SchemaValueType) {
        switch propertyType {
        case .string:
            guard let string = String.fromDatabaseValue(databaseValue) else { return nil }
            self = .string(string)
        case .bool:
            guard let bool = Bool.fromDatabaseValue(databaseValue) else { return nil }
            self = .bool(bool)
        case .int:
            guard let int = Int.fromDatabaseValue(databaseValue) else { return nil }
            self = .int(int)
        case .double:
            guard let double = Double.fromDatabaseValue(databaseValue) else { return nil }
            self = .double(double)
        case .datetime:
            guard let datetimeInt = Int.fromDatabaseValue(databaseValue),
                  let date = DatabaseHelperFunctions.decode(from: datetimeInt) else { return nil }
            self = .datetime(date)
        case .blob:
            guard let data = Data.fromDatabaseValue(databaseValue) else { return nil }
            self = .blob(data)
        }
    }
    
    var databaseValue: DatabaseValue {
        switch self {
        case let .string(string):
            return string.databaseValue
        case let .bool(bool):
            return bool.databaseValue
        case let .int(int):
            return int.databaseValue
        case let .double(double):
            return double.databaseValue
        case let .datetime(datetime):
            return DatabaseHelperFunctions.encode(datetime).databaseValue
        case let .blob(data):
            return data.databaseValue
        }
    }
    
    init(value: Any, propertyType: SchemaValueType, debugInfo: String = "") throws {
        switch propertyType {
        case .double:
            guard let value = value as? Double else {
                throw StringError(description: "Expected Double: \(debugInfo)")
            }
            self = .double(value)
        case .bool:
            guard let value = value as? Bool else {
                throw StringError(description: "Expected Bool: \(debugInfo)")
            }
            self = .bool(value)
        case .int:
            guard let value = value as? Int else {
                throw StringError(description: "Expected Int: \(debugInfo)")
            }
            self = .int(value)
        case .string:
            guard let value = value as? String else {
                throw StringError(description: "Expected String: \(debugInfo)")
            }
            self = .string(value)
        case .datetime:
            guard let value = value as? Int, let date = DatabaseHelperFunctions.decode(from: value) else {
                throw StringError(description: "Expected DateTime (Int with milliseconds since 1970): \(debugInfo)")
            }
            self = .datetime(date)
        case .blob:
            if let data = value as? Data {
                self = .blob(data)
            } else {
                let encodable = AnyEncodable(value)
                do {
                    let data = try JSONEncoder().encode(encodable)
                    self = .blob(data)
                } catch {
                    throw StringError(description: "Expected Data blob: \(debugInfo). \(error)")
                }
            }
        }
    }
    
    func asString() -> String? {
        String(self)
    }
    
    func asDouble() -> Double? {
        Double(self)
    }
    
    func asBool() -> Bool? {
        Bool(self)
    }
    
    func asCGFloat() -> CGFloat? {
        Double(self).flatMap { CGFloat($0) }
    }
}

extension PropertyDatabaseValue: Encodable {
    func encode(to encoder: Encoder) throws {
        var container = encoder.singleValueContainer()
        switch self {
        case let .string(value):
            try container.encode(value)
        case let .bool(value):
            try container.encode(value)
        case let .int(value):
            try container.encode(value)
        case let .double(value):
            try container.encode(value)
        case let .datetime(value):
            try container.encode(value)
        case let .blob(value):
            try container.encode(value)
        }
    }
}


protocol PropertyDatabaseValueConvertible {
    init?(_ value: PropertyDatabaseValue)
}

extension String: PropertyDatabaseValueConvertible {
    init?(_ value: PropertyDatabaseValue) {
        switch value {
        case .string(let value):
            self = value
        case .bool(let value):
            self = value ? "true" : "false"
        case .int(let value):
            self = "\(value)"
        case .double(let value):
            self = value.format("0.###")
        case .datetime(let value):
            let formatter = DateFormatter()
            formatter.dateStyle = .medium
            self = formatter.string(from: value)
        case .blob(let value):
            guard let string = String(data: value, encoding: .utf8) else { return nil }
            self = string
        }
    }
}

extension Bool: PropertyDatabaseValueConvertible {
    init?(_ value: PropertyDatabaseValue) {
            switch value {
            case .string(let value):
                self = value.nilIfBlank != nil && value.lowercased() != "false"
            case .bool(let value):
                self = value
            case .int(let value):
                self = value > 0
            case .double(let value):
                self = value > 0
            default:
                return nil
            }
    }
}

extension Double: PropertyDatabaseValueConvertible {
    init?(_ value: PropertyDatabaseValue) {
        switch value {
        case .string(let value):
            guard let number = Double(value) else { return nil }
            self = number
        case .bool(let value):
            self = value ? 1 : 0
        case .int(let value):
            self = Double(value)
        case .double(let value):
            self = value
        case .datetime(let date):
            self = Double(DatabaseHelperFunctions.encode(date))
        default:
            return nil
        }
    }
}
