//
//  Schema.swift
//  MemriDatabase
//
//  Created by T Brennan on 30/11/20.
//

import Foundation

// MARK: SCHEMA STRUCTS

/// A schema definition. This is used to dynamically enforce supported types and properties
struct Schema {
    /// Supported types in the schema
    var types: [String: SchemaType]
    
    /// Load a Schema struct from the given file URL. Default URL of nil loads `schema.json` from the app bundle
    static func loadFromFile(url: URL? = nil) throws -> Schema {
        func handleError(_ string: String) throws {
            if Constants.isRunningTests {
                throw(StringError(description: string))
            } else {
                print(string)
            }
        }
        
        let fileURL = url ?? Bundle.main.url(forResource: "schema", withExtension: "json")!
        
        let fileData = try Data(contentsOf: fileURL)
        let fileDecoded = try JSONDecoder().decode(SchemaFile.self, from: fileData)
        
        
        let groupedProperties = Dictionary(grouping: fileDecoded.properties) { $0.itemType }
        let groupedEdges = Dictionary(grouping: fileDecoded.edges) { $0.sourceType }
        
        let allTypes = Set(groupedProperties.keys).union(groupedEdges.keys)
        
        // Confirm that all edge target types actually exist in the schema
        let allEdgeTargetTypes = Set(fileDecoded.edges.map { $0.targetType })
        let undefinedEdgeTargetTypes = allEdgeTargetTypes.subtracting(allTypes).subtracting(["Any"])
        if !undefinedEdgeTargetTypes.isEmpty {
            try handleError("Edge target types in schema for types that don't exist: \(undefinedEdgeTargetTypes)")
        }
        
        // Collate the schema into a format we can efficiently traverse
        let types: [String: SchemaType] = Dictionary(uniqueKeysWithValues: allTypes.map { type -> (String, SchemaType) in
            return (type,
                    SchemaType(type: type,
                              propertyTypes: groupedProperties[type]
                                .map { $0.map { ($0.property, $0) } }
                                .map { Dictionary($0, uniquingKeysWith: { a, b in a }) } ?? [:],
                              edgeTypes: groupedEdges[type]
                                .map { $0.map { ($0.edge, $0) } }
                                .map { Dictionary($0, uniquingKeysWith: { a, b in a }) } ?? [:])
            )
        })
        
        return Schema(types: types)
    }
    
    func expectedType(forItemType itemType: String, propertyOrEdgeName: String) -> ResolvedType? {
        if let propertyType = expectedType(forItemType: itemType, propertyName: propertyOrEdgeName) {
            return .property(propertyType)
        } else if let edgeType = expectedTargetType(forItemType: itemType, edgeName: propertyOrEdgeName) {
            return .edge(targetType: edgeType)
        } else {
            return nil
        }
    }
    
    func expectedType(forItemType itemType: String, propertyName: String) -> SchemaValueType? {
        types[itemType]?.propertyTypes[propertyName]?.valueType
    }
    
    func expectedTargetType(forItemType itemType: String, edgeName: String) -> String? {
        types[itemType]?.edgeTypes[edgeName]?.targetType
    }
    
    enum ResolvedType {
        case property(SchemaValueType)
        case edge(targetType: String)
    }
}

/// A schema type definition. This defines the supported properties and their value type
struct SchemaType {
    var type: String
    /// Supported properties for this type in the schema. Value is the value type for the property - this must be a primitive (string, double, dateTime)
    var propertyTypes: [String: SchemaProperty]
    /// Supported properties for this type in the schema. Target type must be another type (eg. Photo, File, etc)
    var edgeTypes: [String: SchemaEdge]
}

/// A schema property definition.
struct SchemaProperty: Codable, Hashable {
    var itemType: String
    var property: String
    var valueType: SchemaValueType
    
    enum CodingKeys: String, CodingKey {
        case itemType = "item_type"
        case property = "property"
        case valueType = "value_type"
    }
}

enum SchemaValueType: String, Codable, Hashable {
    case string
    case bool
    case int
    case double
    case datetime
    case blob
}

/// A schema edge definition.
struct SchemaEdge: Codable {
    var sourceType: String
    var edge: String
    var targetType: String
    var allowsMultiple: Bool
    
    enum CodingKeys: String, CodingKey {
        case sourceType = "source_type"
        case edge = "edge"
        case targetType = "target_type"
    }
    
    init(from decoder: Decoder) throws {
        let container = try decoder.container(keyedBy: CodingKeys.self)
        sourceType = try container.decode(String.self, forKey: .sourceType)
        edge = try container.decode(String.self, forKey: .edge)
        
        let rawTargetType = try container.decode(String.self, forKey: .targetType)
        if rawTargetType.hasSuffix("[]") {
            targetType = String(rawTargetType.dropLast(2))
            allowsMultiple = true
        } else {
            targetType = rawTargetType
            allowsMultiple = false
        }
    }
    
    func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        try container.encode(sourceType, forKey: .sourceType)
        try container.encode(edge, forKey: .edge)
        try container.encode("\(targetType)\(allowsMultiple ? "[]" : "")", forKey: .targetType)
    }
}



// MARK: CODABLE STRUCTS
/// A simple struct used for decoding the schema json.
/// The schema is stored in a diff-friendly format, however the app loads it into memory as dictionaries that can be efficiently queried.
struct SchemaFile: Codable {
    var properties: [SchemaProperty]
    var edges: [SchemaEdge]
}

