//
//  SceneController.swift
//  MemriDatabase
//
//  Created by T Brennan on 14/12/20.
//

import Foundation
import SwiftUI
import Combine
import GRDB

/// The scene controller is specific to a particular `window` of the app. On the iPhone there is usually only one. There may be multiple eg. if multitasking on iPad or multiple windows on mac
class SceneController: ObservableObject {
    let appController: AppController = .shared
    private(set) var topMostContext: ViewContextController?
    let navigationController = MemriUINavigationController()
    
    init() {
        setupObservations()
        
        let oldState = try? appController.databaseController.read {
            try NavigationStack.fetchOne($0)
        }
        if let navStack = oldState, let topView = navStack.state.last {
            navigationStack = navStack
            let context = makeContext(forConfig: topView)
            topMostContext = context
            navigationController.viewControllers = [UIHostingController(rootView: SceneContentView(context: context))]
        } else {
            navigationController.viewControllers = [UIHostingController(rootView: Text("Welcome to Memri"))]
        }
    }
    
    
    
    @Published
    var isContentFullscreen: Bool = false
    
    
    @Published
    var isInEditMode: Bool = false
    
    @Published
    var navigationIsVisible: Bool = false
    var navigationFilterText: String? {
        get { navigationQuery.searchString }
        set { navigationQuery.searchString = newValue?.nilIfBlank }
    }
    
    
    @Published
    var filterPanelIsVisible: Bool = false
    
    private var navigationQuery = DatabaseQueryConfig(itemTypes: ["NavigationItem"], sortProperty: nil) {
        didSet {
            if navigationQuery != oldValue {
                setupObservations()
            }
        }
    }
    
    var navigationItems: [NavigationElement] {
        navigationItemRecords.compactMap { (item) -> NavigationElement? in
            switch item.propertyValue("itemType")?.asString() {
            case "heading":
                guard let title = item.propertyValue("title")?.asString() else {
                    return nil
                }
                return .heading(title)
            case "line":
                return .line
            default:
                guard let title = item.propertyValue("title")?.asString(),
                      let targetViewName = item.propertyValue("sessionName")?.asString() else {
                    return nil
                }
                return .item(.init(name: title, targetViewName: targetViewName))
            }
        }
    }
    
    @Published
    private var navigationItemRecords: [ItemRecord] = []
    private var queryObservation: AnyCancellable?
    /// Sets up a database observation
    private func setupObservations() {
        /// Remove old observation
        queryObservation?.cancel()
        queryObservation = nil

        
        /// Note that the request must remain constant within the observation, hence constructed outside of the tracking (if changed, start a new observation)
        let config = navigationQuery
        let observation = ValueObservation
            .tracking { db in
                try config.executeRequest(db: db)
            }
        
        /// Subscribe to changes in the value, with the first result being reported immediately
        queryObservation = observation.publisher(
            in: appController.databaseController.databasePool,
            scheduling: .immediate)
        .sink(
            receiveCompletion: { completion in },
            receiveValue: { result in
                self.navigationItemRecords = result
            })
    }
    
    private var navigationStack: NavigationStack = NavigationStack() {
            willSet {
                if newValue != navigationStack {
                    objectWillChange.send()
                }
            }
        didSet {
            if navigationStack != oldValue {
                try? appController.databaseController.writeSync {
                    try navigationStack.save($0)
                }
            }
        }
    }
    
    func makeContext(forConfig config: ViewContextHolder, targetItem: ItemRecord? = nil) -> ViewContextController {
        ViewContextController(config: config, focusItem: targetItem, databaseController: appController.databaseController, cvuController: appController.cvuController)
    }
    
    var canNavigateBack: Bool {
        navigationStack.state.count > 1
    }
    
    func navigateBack() {
        guard navigationStack.state.count > 1, let newTopConfig = navigationStack.state[safe: navigationStack.state.endIndex - 2]
        else { return }
        _ = navigationStack.state.popLast()
        
        let context = makeContext(forConfig: newTopConfig)
        topMostContext = context
        
        let vc = UIHostingController(rootView: SceneContentView(context: context))
        navigationController.viewControllers.insert(vc, at: navigationController.viewControllers.endIndex - 1)
        navigationController.popToViewController(vc, animated: true)
    }
    
    func navigateToNewContext(clearStack: Bool = false, animated: Bool = true, viewName: String, inheritDatasource: Bool = false, overrideRenderer: String? = nil, defaultRenderer: String = "list", targetItem: ItemRecord? = nil, overrideUIDs: Set<StringUUID>? = nil, dateRange: DateInterval? = nil, customDefinition: CVUDefinitionContent? = nil, viewArguments: CVUViewArguments? = nil) {
        
        let viewDefinition = appController.cvuController.viewDefinitionFor(viewName: viewName, customDefinition: customDefinition) ?? CVUDefinitionContent()
        
        let newContext = CVUContext(currentItem: nil, selector: nil, viewName: viewName, viewDefinition: viewDefinition, viewArguments: viewArguments)
        
        let rendererName = overrideRenderer ?? { () -> String in
            return viewDefinition.propertyResolver(context: newContext, lookup: CVULookupController(), db: appController.databaseController).string("defaultRenderer") ?? defaultRenderer
        }()
        
        let datasource = viewDefinition.definitions.first(where: { $0.type == .datasource })
        let datasourceResolver = datasource?.parsed.propertyResolver(context: newContext, lookup: CVULookupController(), db: appController.databaseController)
        let uidList = overrideUIDs ?? datasourceResolver.map {
            Set($0.stringArray("uids").map { StringUUID(uid: $0) })
        }
        let filterDef = datasourceResolver?.subdefinition("filter")
        let edgeTargets = filterDef?.subdefinition("edgeTargets")
        let edgeTargetConditions = edgeTargets?.properties.keys.compactMap { key -> DatabaseQueryCondition? in
            guard let target = edgeTargets?.string(key) else { return nil }
            return .edgeHasTarget(DatabaseQueryCondition.EdgeHasTarget(edgeName: key, targetUID: StringUUID(uid: target)))
        } ?? []
        
        var queryConfig = inheritDatasource ? (topMostContext?.config.query ?? DatabaseQueryConfig()) : DatabaseQueryConfig()
        if let itemTypes = datasourceResolver?.stringArray("query") ?? targetItem.map({ [$0.type] })  {
            queryConfig.itemTypes = itemTypes
        }
        if let itemUIDs = uidList {
            queryConfig.itemUIDs = itemUIDs
        }
        if let sortProperty = datasourceResolver?.string("sortProperty") {
            queryConfig.sortProperty = sortProperty
        }
        if let dateModifiedRange = dateRange {
            queryConfig.dateModifiedAfter = dateModifiedRange.start
            queryConfig.dateModifiedBefore = dateModifiedRange.end
        }
        if !edgeTargetConditions.isEmpty {
            queryConfig.conditions = edgeTargetConditions
        }
        
        let config = ViewContext(viewName: viewName, rendererName: rendererName, viewDefinition: viewDefinition, query: queryConfig, viewArguments: viewArguments)
        let holder = ViewContextHolder(config)
        
        let newViewContextController = makeContext(forConfig: holder, targetItem: targetItem)
        topMostContext = newViewContextController
        if clearStack {
            navigationStack.state = [holder]
        } else {
            navigationStack.state.append(holder)
        }
        navigationController.setViewControllers([UIHostingController(rootView: SceneContentView(context: newViewContextController))], animated: animated)
    }
}
