//
//  SyncController.swift
//  MemriDatabase
//
//  Created by T Brennan on 1/12/20.
//

import Foundation
import GRDB

enum SyncState: String, Codable, Hashable, DatabaseValueConvertible {
    case create
    case update
    case noChanges
    case failed
}

class SyncController {
    let databaseController: DatabaseController
    
    init(databaseController: DatabaseController) {
        self.databaseController = databaseController
    }
    
    func syncUpload(maxItems: Int = 100) throws {
        /// NOTE: Implementing this will depend on the new Pod API once the dynamic schema is implemented.
        /// The `syncPayload` above can be passed to the bulk-action API call. Then use the response to determine if the items where successfully synced (if so, mark them as synced so they are not synced again until changed)
        
        /// Retrieve connection details
//        guard let connectionConfig = AppController.shared.podConnectionConfig
//          else { throw StringError(description: "No pod connection config") }
        
        /// Construct the JSON to send to the pod
//        let syncPayload = try makeSyncUploadData(maxItems: maxItems)
        
        /// Start network task
//        let request = PodAPIStandardRequest.bulkAction(payload: syncPayload)
//        let networkCall = try request.execute(connectionConfig: connectionConfig)
        
//        networkCall.sink { response in
            /// Handle network task completion (eg. mark items as synced or failed)
            
            /// Schedule another sync operation until all batches are complete
//        }
    }
    
    func makeSyncUploadData(maxItems: Int = 100) throws -> PodAPIPayload.BulkAction {
        return try databaseController.databasePool.read { (db) in
            
            /// Select the items to sync, giving priority to those marked as `syncHasPriority`
            let itemsForSync = try ItemRecord.order(ItemRecord.Columns.syncHasPriority.desc).limit(maxItems).fetchAll(db).filter { ![SyncState.noChanges, SyncState.failed].contains($0.syncState) }
            
            /// Construct a JSON-compatible representation for the items
            let syncItems = itemsForSync.map { itemRecord -> SyncUploadItem in
                let properties = itemRecord.syncDict(db: db, schema: databaseController.schema)
                return SyncUploadItem(operation: itemRecord.syncState, properties: properties)
            }
            
            /// Group by sync operation. eg. create, update
            let itemsByOperation = Dictionary(grouping: syncItems) { $0.operation }.mapValues { $0.map { $0.properties } }
            
            let bulkAction = PodAPIPayload.BulkAction(
                createItems: itemsByOperation[.create] ?? [],
                updateItems: itemsByOperation[.update] ?? [],
                deleteItems: [],
                createEdges: [],
                updateEdges: [],
                deleteEdges: [])
            return bulkAction
        }
    }
    
    func syncDownload(json: Data, schema: Schema) throws {
        let dict = try JSONDecoder().decode([String: [[String: AnyDecodable]]].self, from: json)
        let createItemDicts = dict[SyncState.create.rawValue] ?? []
        let updateItemsDicts = dict[SyncState.update.rawValue] ?? []
        
        let createItems = createItemDicts.compactMap { SyncDownloadItem(fromSyncItemDict: $0, schema: schema) }
        let updateItems = updateItemsDicts.compactMap { SyncDownloadItem(fromSyncItemDict: $0, schema: schema) }
        
        print("Sync download test: Create \(createItems.count) items, Update \(updateItems.count) items")
    }
}

struct SyncUploadItem {
    var operation: SyncState
    var properties: [String: AnyEncodable]
}

struct SyncDownloadItem {
    var item: ItemRecord
    var properties: [ItemPropertyRecord]
    
    init?(fromSyncItemDict dict: [String: AnyDecodable], schema: Schema) {
        guard let item = try? ItemRecord(fromSyncDict: dict) else { return nil }
        let properties = dict
            .compactMap { (propertyName, decodableValue) -> ItemPropertyRecord? in
                guard !ItemRecord.intrinsicProperties.contains(propertyName) else { return nil }
                guard let expectedType = schema.types[item.type]?.propertyTypes[propertyName] else {
                    return nil
                }
                do {
                    let propertyValue = decodableValue.value
                    let result = try PropertyDatabaseValue(value: propertyValue, propertyType: expectedType.valueType, debugInfo: "\(item.type).\(propertyName)")
                    return ItemPropertyRecord(itemUID: item.uid, name: propertyName, value: result)
                }
                catch {
                    return nil
                }
            }
        self.item = item
        self.properties = properties
    }
}
