//
// CVU_Image.swift
// Copyright © 2020 memri. All rights reserved.

import Foundation
import SwiftUI

/// A CVU element for displaying an image
/// - Use the `image` property to display an Image item
/// - Use the `bundleImage` property to display an icon pre-packaged with the Memri App
/// - Use the `systemName` property to display an iOS SFSymbols (system) icon
struct CVU_Image: View {
    var nodeResolver: CVUUINodeResolver

    var fileImageURL: URL? {
        guard let imageURI = nodeResolver.propertyResolver.fileUID(key: "image")
        else { return nil }
        return FileStorageController.getURLForFile(withUUID: imageURI)
    }

    var bundleImage: UIImage? {
        guard let imageName = nodeResolver.propertyResolver.string("bundleImage"),
              let image = UIImage(named: imageName)
        else {
            return nil
        }
        return image
    }

    @ViewBuilder
    var body: some View {
        if let imageURL = fileImageURL {
            MemriImageView(imageURL: imageURL, fitContent: nodeResolver.propertyResolver.sizingMode() == .fit)
                .if(nodeResolver.propertyResolver.sizingMode() == .fit) {
                    $0.aspectRatio(
                        MemriImageView.getAspectRatio(of: imageURL) ?? 1,
                        contentMode: .fit
                    )
                }
        }
        else if let bundleImage = bundleImage {
            Image(uiImage: bundleImage)
                .resizable()
                .if(nodeResolver.propertyResolver.sizingMode() == .fit) { $0.aspectRatio(contentMode: .fit) }
        }
        else if let iconName = nodeResolver.propertyResolver.string("systemName") {
            Image(systemName: iconName)
                .renderingMode(.template)
                .if(nodeResolver.propertyResolver.bool("resizable", defaultValue: false)) { $0.resizable() }
                .if(nodeResolver.propertyResolver.sizingMode() == .fit) { $0.aspectRatio(contentMode: .fit) }
        }
        else {
            Image(systemName: "questionmark")
                .renderingMode(.template)
                .aspectRatio(contentMode: .fit)
                .foregroundColor(Color(.secondaryLabel))
                .frame(maxWidth: .infinity, maxHeight: .infinity)
                .layoutPriority(-1)
        }
    }
}
