//
// CVU_Map.swift
// Copyright © 2020 memri. All rights reserved.

import Foundation
import SwiftUI
import CoreLocation

/// A CVU element for displaying a Map
struct CVU_Map: View {
    var nodeResolver: CVUUINodeResolver

    var content: String? {
        nodeResolver.propertyResolver.string("text")?.nilIfBlank
    }

    var locationResolver: (ItemRecord) -> [ItemRecord] {{ _ in
        self.nodeResolver.propertyResolver.items("location")
    }}
    var addressResolver: (ItemRecord) -> [ItemRecord] {{ _ in
        self.nodeResolver.propertyResolver.items("address")
    }}

    var labelResolver: (ItemRecord) -> String? {{ _ in
        self.nodeResolver.propertyResolver.string("label")
    }}

    var config: MapViewConfig {
        MapViewConfig(dataItems: nodeResolver.propertyResolver.context.currentItem.map { [$0] } ?? [],
                      locationResolver: locationResolver,
                      addressResolver: addressResolver,
                      labelResolver: labelResolver,
                      moveable: nodeResolver.propertyResolver.bool("moveable", defaultValue: true))
    }

    var body: some View {
        MapView(config: config)
            .background(Color(.secondarySystemBackground))
    }
}

import CoreLocation

struct MapViewConfig {
    var dataItems: [ItemRecord] = []
    var locationResolver: (ItemRecord) -> [ItemRecord]
    var addressResolver: (ItemRecord) -> [ItemRecord]
    var labelResolver: (ItemRecord) -> String?
    var maxInitialZoom: Double = 16
    var moveable: Bool = true
//    var mapStyle: MapStyle = .street

    var onPress: ((ItemRecord) -> Void)?

    @Environment(\.colorScheme) var colorScheme
}

import Combine
class MapModel {
    var didChange = PassthroughSubject<Void, Never>()

    init() {
        updateQueue.throttle(for: .seconds(0.1), scheduler: RunLoop.main, latest: true)
            .sink { [weak self] in
                self?._updateModel()
            }
            .store(in: &cancellableBag)
    }

    var dataItems: [ItemRecord] = [] {
        didSet {
            if dataItems != oldValue {
                updateModel()
            }
        }
    }

    private(set) var items: [MapItem] = [] {
        didSet {
            if items != oldValue {
                didChange.send()
            }
        }
    }

    var locationResolver: ((ItemRecord) -> [ItemRecord])?
    var addressResolver: ((ItemRecord) -> [ItemRecord])?
    var labelResolver: ((ItemRecord) -> String?)?

    func updateModel() {
        updateQueue.send()
    }

    var cancellableBag = Set<AnyCancellable>()
    let updateQueue = PassthroughSubject<Void, Never>()

    func _updateModel() {
        let newItems = dataItems.flatMap { dataItem -> [MapItem] in
            let locations: [CLLocation] = resolveItem(dataItem: dataItem)
            let labelString: String = labelResolver?(dataItem) ?? ""
            return locations.map {
                return MapItem(label: labelString, coordinate: $0.coordinate, dataItem: dataItem)
            }
        }
        items = newItems
    }

    func resolveItem(dataItem: ItemRecord) -> [CLLocation] {
        var clLocations: [CLLocation] = []
        
        if let locations = locationResolver?(dataItem).compactMap({ (item) -> CLLocation? in
            guard let latitude = item.propertyValue("latitude")?.asDouble(),
                  let longitude = item.propertyValue("longitude")?.asDouble() else {
                return nil
            }
            return CLLocation(latitude: latitude, longitude: longitude)
        }) {
            clLocations.append(contentsOf: locations)
        }

        let addresses = addressResolver?(dataItem) ?? []
        let resolvedLocations = addresses.compactMap { self.lookupAddress($0) }
        clLocations.append(contentsOf: resolvedLocations)

        return clLocations
    }

    var addressLookupCancellables: [ItemRecord: AnyCancellable] = [:]

    func lookupAddress(_ address: ItemRecord) -> CLLocation? {
        if let location = MapHelper.shared.getLocationForAddress(address: address) {
            return location
        } else {
            let lookupPublisher = MapHelper.shared.lookupLocationForAddress(address: address)
            addressLookupCancellables[address] = lookupPublisher.sink { [weak self] _ in self?.updateModel() }
            return nil
        }
    }

    struct MapItem: Equatable {
        var label: String
        var coordinate: CLLocationCoordinate2D
        var dataItem: ItemRecord?
    }
}

extension CLLocationCoordinate2D: Equatable {
    public static func == (lhs: Self, rhs: Self) -> Bool {
        lhs.latitude == rhs.latitude && lhs.longitude == rhs.longitude
    }
}
