//
// MapHelper.swift
// Copyright © 2020 memri. All rights reserved.

import Combine
import CoreLocation
import Foundation
import MapKit

class MapHelper {
    static var shared = MapHelper()

    private init() {}
    
    private func lookupAddress(string: String) -> Future<[MKMapItem], Never> {
        Future { promise in
            let request = MKLocalSearch.Request()
            request.naturalLanguageQuery = string
            MKLocalSearch(request: request).start { response, error in
                guard let locations = response?.mapItems, !locations.isEmpty else {
                    if let error = error {
                        // Log error
                        print(error.localizedDescription)
                    }
                    promise(.success([])) // Rather than propogating the error we return 0 results
                    return
                }
                promise(.success(locations))
            }
        }
    }

    func getLocationForAddress(address: ItemRecord, db: DatabaseController = AppController.shared.databaseController)
        -> CLLocation?
    {
        // Check if the address holds a valid location
        if let location = address.edgeItem("location"),
           let latitude = location.propertyValue("latitude")?.asDouble(),
           let longitude = location.propertyValue("longitude")?.asDouble() {
            return CLLocation(latitude: latitude, longitude: longitude)
        }
        return nil
    }
    
    func lookupLocationForAddress(address: ItemRecord, db: DatabaseController = AppController.shared.databaseController)
        -> AnyPublisher<CLLocation?, Never>
    {
        
        // Make new lookup
        let inclusions = ["street", "city", "state"]
        let addressString = inclusions.compactMap { address.propertyValue($0)?.asString() }.joined(separator: " ")
        let lookupHash = addressString.hashValue
        
        // Check lookups in progress
        if let knownLocation = addressLookupResults[lookupHash] {
            // Successful lookup already completed
            return Just(knownLocation).eraseToAnyPublisher()
        }
        else if let existingLookup = addressLookupPublishers[lookupHash] {
            // Lookup already in progress
            return existingLookup
        }
        

        // Check if the address holds a valid location
        if let location = address.edgeItem("location"),
           let latitude = location.propertyValue("latitude")?.asDouble(),
           let longitude = location.propertyValue("longitude")?.asDouble()
        {
            let clLocation = CLLocation(latitude: latitude, longitude: longitude)
            if let oldLookupHash = address.propertyValue("locationAutoLookupHash")?.asString() {
                // This was an automatic lookup - check it's still current
                if oldLookupHash == String(lookupHash) { return Just(clLocation).eraseToAnyPublisher() }
            }
            else {
                // No old lookup hash, this location is user-defined
                return Just(clLocation).eraseToAnyPublisher()
            }
        }

        
        // Make new lookup
        let lookup = lookupAddress(string: addressString)
            .map { (mapItem) -> CLLocation? in
                // Find the first placemark that has a location value
                mapItem.first { (mapItem) -> Bool in
                    mapItem.placemark.location != nil
                }?.placemark.location
            }
            .flatMap {[weak self] location in
                Future<CLLocation?, Never> { (promise) in
                    let onCompletion = { [weak self] in
                        self?.addressLookupResults[lookupHash] = location
                        self?.addressLookupPublishers[lookupHash] = nil
                        promise(.success(location))
                    }
                    if let location = location {
                        // Update the address with the location (avoid future lookups)
                        let locationItem = ItemRecord(type: "Location")
                        let locationItemProperties = [
                            ItemPropertyRecord(itemUID: locationItem.uid, name: "latitude", value: .double(location.coordinate.latitude)),
                            ItemPropertyRecord(itemUID: locationItem.uid, name: "longitude", value: .double(location.coordinate.longitude))
                        ]
                        let locationEdge = ItemEdgeRecord(sourceUID: address.uid, name: "location", targetUID: locationItem.uid)
                        
                        AppController.shared.databaseController.write { db in
                            try locationItem.save(db)
                            try locationItemProperties.forEach { try $0.save(db) }
                            try locationEdge.save(db)
                            onCompletion()
                        }
                    } else {
                        onCompletion()
                    }
                }
            }
            .multicast(subject: PassthroughSubject())
            .autoconnect()
            .eraseToAnyPublisher()
        addressLookupPublishers[lookupHash] = lookup

        return lookup
    }

    var cancellableBag = Set<AnyCancellable>()

    var addressLookupResults: [Int: CLLocation] = [:]
    var addressLookupPublishers: [Int: AnyPublisher<CLLocation?, Never>] = [:]
}
