//
// NavigationView.swift
// Copyright © 2020 memri. All rights reserved.

import ASCollectionView
import SwiftUI

/// This view is the main  NavigationPane. It lists NavigationItems and provides search functionality for this list.
struct NavigationPaneView: View {
    @ObservedObject var sceneController: SceneController

    @ObservedObject var keyboardResponder = KeyboardResponder.shared

    @State var showSettings: Bool = false

    var body: some View {
        VStack(spacing: 0) {
            HStack(spacing: 20) {
                Button(action: {
                    self.showSettings = true
                }) {
                    Image(systemName: "gear")
                        .font(Font.system(size: 22, weight: .semibold))
                        .foregroundColor(Color(hex: "#d9d2e9"))
                }
//                .sheet(isPresented: self.$showSettings) {
//                    SettingsPane().environmentObject(self.context)
//                }

                MemriTextField(
                    value: $sceneController.navigationFilterText,
                    placeholder: "Search",
                    textColor: UIColor(hex: "#8a66bc"),
                    tintColor: UIColor.white,
                    clearButtonMode: .always,
                    showPrevNextButtons: false
                )
                .layoutPriority(-1)
                .padding(5)
                .padding(.horizontal, 5)
                .accentColor(.white)
                .background(Color.black.opacity(0.4))
                .cornerRadius(5)

            }
            .padding(.top, 40)
            .padding(.horizontal, 20)
            .frame(minHeight: 95)
            .background(Color("MemriUI-purpleBackSecondary"))

            ASTableView(section:
                            ASSection(id: 0, data: sceneController.navigationItems,
                          dataID: \.self) { navItem, _ in
                                switch navItem {
                                case .item(let item):
                                    NavigationItemView(item: item, sceneController: sceneController)
                                case .heading(let title):
                                    NavigationHeadingView(title: title)
                                case .line:
                                    NavigationLineView()
                                }
                })
                .separatorsEnabled(false)
                .contentInsets(UIEdgeInsets(top: 15, left: 0, bottom: 0, right: 0))

        }
        .frame(maxWidth: .infinity, alignment: .leading)
        .background(Color("MemriUI-purpleBack"))
    }
}

enum NavigationElement: Hashable {
    case item(Item)
    case heading(String)
    case line
    
    struct Item: Hashable {
        var name: String
        var targetViewName: String
    }
}


struct NavigationItemView: View {
    var item: NavigationElement.Item
    var sceneController: SceneController

    var body: some View {
        Button(action: {
            sceneController.navigateToNewContext(clearStack: true, animated: false, viewName: item.targetViewName)
            withAnimation {
                sceneController.navigationIsVisible = false
            }
        }) {
            Text(item.name.localizedCapitalized)
                .font(.system(size: 18, weight: .regular))
                .padding(.vertical, 10)
                .padding(.horizontal, 35)
                .foregroundColor(Color.white.opacity(0.7))
                .frame(maxWidth: .infinity, alignment: .leading)
                .contentShape(Rectangle())
        }
        .buttonStyle(NavigationButtonStyle())
    }
}

struct NavigationButtonStyle: ButtonStyle {
    func makeBody(configuration: Configuration) -> some View {
        configuration.label
            .background(
                configuration.isPressed ? Color.white.opacity(0.15) : .clear
            )
    }
}

struct NavigationHeadingView: View {
    var title: String?

    var body: some View {
        HStack {
            Text((title ?? "").uppercased())
                .font(.system(size: 18, weight: .bold))
                .padding(.horizontal, 20)
                .padding(.vertical, 8)
                .foregroundColor(Color(hex: "#8c73af"))
            Spacer()
        }
    }
}

struct NavigationLineView: View {
    var body: some View {
        VStack {
            Divider().background(Color(.black))
        }.padding(.horizontal, 50)
    }
}

//struct Navigation_Previews: PreviewProvider {
//    static var previews: some View {
//        NavigationView()
//    }
//}
