//
// CalendarRenderer.swift
// Copyright © 2020 memri. All rights reserved.

import ASCollectionView
import SwiftUI

/// The calendar renderer
/// This presents the data in a month-style calendar view
/// Dots are used to represent days on which items fall
/// Pressing on a day will show a timeline view focused on that day
struct CalendarRendererView: View {
    @EnvironmentObject var sceneController: SceneController
    @ObservedObject var viewContext: ViewContextController

    @State var scrollPosition: ASCollectionViewScrollPosition? = .bottom
    
    var calendarHelper = CalendarHelper()

    func calculateModel() -> CalendarCalculations {
        CalendarCalculations(calendarHelper: calendarHelper,
                             data: viewContext.items,
                             dateResolver: { item -> Date? in
                                viewContext.nodePropertyResolver(for: item)?.dateTime("dateTime") ?? item.dateModified
                            })
    }
    
    var primaryColor: CVUColor { viewContext.rendererDefinitionPropertyResolver.color() ?? CVUColor.system(.red) }

    var body: some View {
        let backgroundColor = viewContext.rendererDefinitionPropertyResolver.backgroundColor ?? CVUColor.system(.systemBackground)
        
        let model = calculateModel()
        
        
        VStack(spacing: 0) {
            HStack(spacing: 0) {
                ForEach(calendarHelper.daysInWeek, id: \.self) { dayString in
                    Text(dayString)
                        .font(.headline)
                        .frame(maxWidth: .infinity)
                }
            }
            .padding(.horizontal, 20)
            .background(Color.gray.opacity(0.2))
            ASCollectionView(sections: sections(withCalcs: model))
                .scrollPositionSetter($scrollPosition)
                .layout(ASCollectionLayout(scrollDirection: .vertical, interSectionSpacing: 4) {
                    ASCollectionLayoutSection { (_) -> NSCollectionLayoutSection in
                        let columns = 7

                        let itemSize = NSCollectionLayoutDimension.absolute(55)

                        let itemLayoutSize = NSCollectionLayoutSize(
                            widthDimension: .fractionalWidth(1.0),
                            heightDimension: itemSize
                        )
                        let groupSize = NSCollectionLayoutSize(
                            widthDimension: .fractionalWidth(1.0),
                            heightDimension: itemSize
                        )
                        let supplementarySize = NSCollectionLayoutSize(
                            widthDimension: .fractionalWidth(1.0),
                            heightDimension: .absolute(30)
                        )

                        let item = NSCollectionLayoutItem(layoutSize: itemLayoutSize)

                        let group = NSCollectionLayoutGroup.horizontal(
                            layoutSize: groupSize,
                            subitem: item,
                            count: columns
                        )

                        let section = NSCollectionLayoutSection(group: group)
                        section.interGroupSpacing = 0
                        section.contentInsets = .init(top: 0, leading: 20, bottom: 0, trailing: 20)
                        section
                            .visibleItemsInvalidationHandler = { _, _, _ in
                            } // If this isn't defined, there is a bug in UICVCompositional Layout that will fail to update sizes of cells

                        let headerSupplementary = NSCollectionLayoutBoundarySupplementaryItem(
                            layoutSize: supplementarySize,
                            elementKind: UICollectionView.elementKindSectionHeader,
                            alignment: .top
                        )
                        let footerSupplementary = NSCollectionLayoutBoundarySupplementaryItem(
                            layoutSize: supplementarySize,
                            elementKind: UICollectionView.elementKindSectionFooter,
                            alignment: .bottom
                        )
                        section.boundarySupplementaryItems = [
                            headerSupplementary,
                            footerSupplementary,
                        ]
                        return section
                    }
                })
                .contentInsets(UIEdgeInsets(top: 10, left: 0, bottom: 10, right: 0))
                .alwaysBounceVertical()
        }
        .background(backgroundColor.color)
    }

    func sections(withCalcs calcs: CalendarCalculations) -> [ASSection<Date>] {
        calendarHelper.getMonths(from: calcs.start, to: calcs.end)
            .map { section(forMonth: $0, withCalcs: calcs) }
    }

    func section(forMonth month: Date, withCalcs calcs: CalendarCalculations) -> ASSection<Date> {
        let days = calendarHelper.getPaddedDays(forMonth: month)
        return ASSection(
            id: month,
            data: days,
            dataID: \.self,
            selectionMode: .selectSingle { index in
                guard let day = days[safe: index].flatMap({ $0 }),
                      calcs.hasItemOnDay(day)
                else { return }
                CVUAction_OpenView(viewName: "calendarDayView", renderer: "timeline", dateRange: calendarHelper.wholeDay(for: day)).execute(sceneController: sceneController, context: viewContext.getCVUContext())
            }
        ) { day, cellContext in
            Group {
                day.map { day in
                    VStack(spacing: 0) {
                        Spacer()
                        Text(calendarHelper.dayString(for: day))
                            .foregroundColor(
                                calendarHelper.isToday(day) ? primaryColor.color : Color(.label))

                        HStack(spacing: 0) {
                            Circle()
                                .fill(calcs.itemsOnDay(day).isEmpty ? .clear : primaryColor.color)
                                .frame(width: 10, height: 10)
                                .padding(4)
                            if calcs.itemsOnDay(day).count > 1 {
                                Text("×\(calcs.itemsOnDay(day).count)")
                                    .font(Font.caption.bold())
                                    .foregroundColor(primaryColor.color)
                                    .fixedSize()
                            }
                        }

                        Spacer()
                        Divider()
                    }
                    .background(cellContext
                        .isHighlighted ? Color(.darkGray).opacity(0.3) : .clear)
                }
            }
        }
        .sectionHeader {
            Text(calendarHelper.monthYearString(for: month))
                .font(.headline)
                .frame(maxWidth: .infinity, alignment: .leading)
        }
    }
}
