//
// ChartRenderer.swift
// Copyright © 2020 memri. All rights reserved.

import Combine
import Foundation
import SwiftUI

/// The chart renderer.
/// This renderer displays the data in a chart (eg. line, bar, pie)
struct ChartRendererView: View {
    @EnvironmentObject var sceneController: SceneController
    @ObservedObject var viewContext: ViewContextController

    var missingDataView: some View {
        Text("You need to define x/y axes in CVU").bold().frame(maxWidth: .infinity, maxHeight: .infinity)
    }
    
    var backgroundColor: CVUColor {
        viewContext.rendererDefinitionPropertyResolver.backgroundColor ?? CVUColor.system(.systemBackground)
    }
    
    var chartTitle: String? {
        viewContext.rendererDefinitionPropertyResolver.string("title")
    }
    
    var chartSubtitle: String? {
        viewContext.rendererDefinitionPropertyResolver.string("subtitle")
    }
    
    var chartType: ChartType {
        viewContext.rendererDefinitionPropertyResolver.string("chartType").flatMap(ChartType.init) ?? .bar
    }
    
    var primaryColor: CVUColor {
        viewContext.rendererDefinitionPropertyResolver.color() ?? .system(.blue)
    }
    
    var lineWidth: CGFloat {
        viewContext.rendererDefinitionPropertyResolver.cgFloat("lineWidth") ?? 0
    }
    
    var yAxisStartAtZero: Bool {
        viewContext.rendererDefinitionPropertyResolver.bool("yAxisStartAtZero", defaultValue: false)
    }
    
    var hideGridlines: Bool {
        viewContext.rendererDefinitionPropertyResolver.bool("hideGridlines", defaultValue: false)
    }
    
    var barLabelFont: CVUFont {
        viewContext.rendererDefinitionPropertyResolver.font(for: "barLabelFont", baseFont: CVUFont(size: 13))
    }
    
    var showValueLabels: Bool {
        viewContext.rendererDefinitionPropertyResolver.bool("yAxisStartAtZero", defaultValue: true)
    }
    
    var valueLabelFont: CVUFont {
        viewContext.rendererDefinitionPropertyResolver.font(for: "valueLabelFont", baseFont: CVUFont(size: 14))
    }
    
    func makeBarChartModel() -> ChartModelLabelledY? {
        let resolver = viewContext.rendererDefinitionPropertyResolver
    
            let data = ChartHelper.generateLabelledYChartSetFromItems(
                viewContext.items,
                labelExpression: { resolver.replacingItem($0).string("label") },
                yAxis: { resolver.replacingItem($0).double("yAxis") },
                yAxisLabel: { resolver.replacingItem($0).string("yAxisLabel") }
            )
    
            return ChartModelLabelledY(
                sets: [data],
                hideGridLines: hideGridlines,
                forceMinYOfZero: yAxisStartAtZero,
                primaryColor: primaryColor,
                barLabelFont: barLabelFont.uiFont,
                showValueLabels: showValueLabels,
                valueLabelFont: valueLabelFont.uiFont
            )
    }
    
    func makeLineChartModel() -> LineChartModel? {
        let resolver = viewContext.rendererDefinitionPropertyResolver
        
        let data = ChartHelper.generateXYChartSetFromItems(
            viewContext.items,
            xAxis: { resolver.replacingItem($0).double("xAxis") },
            yAxis: { resolver.replacingItem($0).double("yAxis") },
            labelExpression: {
                resolver.replacingItem($0).string("label")
            }
        )
        
        return LineChartModel(
            sets: [data],
            lineWidth: lineWidth,
            hideGridLines: hideGridlines,
            forceMinYOfZero: yAxisStartAtZero,
            primaryColor: primaryColor,
            showValueLabels: showValueLabels,
            valueLabelFont: valueLabelFont.uiFont
        )
    }
    
    // MARK: Views

    var chartTitleView: some View {
        VStack {
            chartTitle.map {
                Text($0)
                    .font(.title)
            }
            chartSubtitle.map {
                Text($0)
                    .foregroundColor(Color(.secondaryLabel))
                    .font(.body)
            }
        }
    }

    var chartView: AnyView {
        switch chartType {
        case .bar:
            guard let model = makeBarChartModel() else {
                return missingDataView.eraseToAnyView()
            }

            return VStack(spacing: 0) {
                chartTitleView
                BarChartSwiftUIView(
                    model: model,
                    onPress: onPress(index:)
                )
            }
            .padding(10)
            .eraseToAnyView()
        case .line:
            guard let model = makeLineChartModel() else {
                return missingDataView.eraseToAnyView()
            }

            return VStack(spacing: 0) {
                chartTitleView
                LineChartSwiftUIView(
                    model: model,
                    onPress: onPress(index:)
                )
            }
            .padding(10)
            .eraseToAnyView()
        case .pie:
            guard let model = makeBarChartModel() else {
                return missingDataView.eraseToAnyView()
            }

            return VStack(spacing: 0) {
                chartTitleView
                PieChartSwiftUIView(
                    model: model,
                    holeColor: backgroundColor,
                    onPress: onPress(index:)
                )
            }
            .padding(10)
            .eraseToAnyView()
        }
    }
    
    func onPress(index: Int) {
        if let item = viewContext.items[safe: index],
           let press = viewContext.nodePropertyResolver(for: item)?.action(key: "onPress") {
            press.execute(sceneController: sceneController, context: viewContext.getCVUContext(forItem: item))
        }
    }

    var body: some View {
        chartView
            .frame(maxWidth: .infinity, maxHeight: .infinity)
            .background(backgroundColor.color)
    }
}
