//
// GridRenderer.swift
// Copyright © 2020 memri. All rights reserved.

import ASCollectionView
import SwiftUI


/// The grid renderer
/// This presents the data in a grid (aka collection view)
struct GridRendererView: View {
    @EnvironmentObject var sceneController: SceneController
    @ObservedObject var viewContext: ViewContextController
 
    var scrollDirection: UICollectionView.ScrollDirection {
        switch viewContext.rendererDefinitionPropertyResolver.string("scrollDirection") {
        case "horizontal": return .horizontal
        default: return .vertical
        }
    }
    
    var backgroundColor: CVUColor {
        viewContext.rendererDefinitionPropertyResolver.backgroundColor ?? CVUColor.system(.systemBackground)
    }

    var layout: GridRendererLayout {
        switch viewContext.rendererDefinitionPropertyResolver.string("layout")?.lowercased() {
        case "photogrid":
            return GridRendererLayout_PhotoGrid(renderConfigResolver: viewContext.rendererDefinitionPropertyResolver)
        case "waterfall":
            return GridRendererLayout_Waterfall(renderConfigResolver: viewContext.rendererDefinitionPropertyResolver)
        default:
            return GridRendererLayout_Grid(renderConfigResolver: viewContext.rendererDefinitionPropertyResolver)
        }
    }

    var section: ASCollectionViewSection<Int> {
        ASCollectionViewSection(id: 0,
                                data: viewContext.items,
                                selectionMode: selectionMode,
                                contextMenuProvider: contextMenuProvider)
        { dataItem, state in
            ZStack(alignment: .bottomTrailing) {
                viewContext.render(item: dataItem)

                if sceneController.isInEditMode && !state.isSelected {
                    Color.white.opacity(0.15)
                }
                if state.isSelected {
                    ZStack {
                        Circle().fill(Color.blue)
                        Circle().strokeBorder(Color.white, lineWidth: 2)
                        Image(systemName: "checkmark")
                            .font(.system(size: 15, weight: .bold))
                            .foregroundColor(.white)
                    }
                    .frame(width: 30, height: 30)
                    .padding(10)
                }
            }
        }
    }


    var body: some View {
        VStack {
            if viewContext.hasItems {
                ASCollectionView(editMode: sceneController.isInEditMode, section: section)
                    .layout(self.layout.layout)
                    .alwaysBounceHorizontal(scrollDirection == .horizontal)
                    .alwaysBounceVertical(scrollDirection == .vertical)
                    .if(layout.layoutName == "waterfall") {
                        $0.customDelegate(GridRendererLayout_Waterfall_Delegate.init)
                    }
                    .background(backgroundColor.color)
                    .id(self.layout.layoutStateID) // Reload if layout state changed
            }
            else {
                HStack(alignment: .top) {
                    Spacer()
                    Text("No results")
                        .multilineTextAlignment(.center)
                        .font(.system(size: 16, weight: .regular, design: .default))
                        .opacity(0.7)
                    Spacer()
                }
                .padding(.all, 30)
                .padding(.top, 40)
                Spacer()
            }
        }
    }
    
    
    var selectionMode: ASSectionSelectionMode {
        if sceneController.isInEditMode {
            return .selectMultiple(viewContext.selectedIndicesBinding)
        }
        else {
            return .selectSingle { index in
                if let item = viewContext.items[safe: index],
                   let press = viewContext.nodePropertyResolver(for: item)?.action(key: "onPress") {
                    press.execute(sceneController: sceneController, context: viewContext.getCVUContext(forItem: item))
                }
            }
        }
    }
    
    func contextMenuProvider(index: Int, item: ItemRecord) -> UIContextMenuConfiguration? {
        nil
//        let children: [UIMenuElement] = config.contextMenuActions.map { [weak context] action in
//            UIAction(title: action.getString("title"),
//                     image: nil) { [weak context] _ in
//                context?.executeAction(action, with: item)
//            }
//        }
//        guard !children.isEmpty else { return nil }
//        return UIContextMenuConfiguration(identifier: nil, previewProvider: nil) { (_) -> UIMenu? in
//            UIMenu(title: "", children: children)
//        }
    }
}

protocol GridRendererLayout {
    init(renderConfigResolver: CVUPropertyResolver)
    var layout: ASCollectionLayout<Int> { get }
    
    /// The name of this layout type
    var layoutName: String { get }
    
    /// A value that changes when the layout needs to be reloaded
    var layoutStateID: String { get }
}
