//
// LabelAnnotationRenderer.swift
// Copyright © 2020 memri. All rights reserved.

import Foundation
import SwiftUI

/// The Label Annotation renderer
/// This renderer can be used to annotate items with Labels
/// - Set the `labelType` to a unique ID representing the type of labels
/// - Set the `labelOptions` property to an array of Strings (representing the options)
/// - You may also provide a `labelOptionIcons` property with an array of icon names (in matching order with `labelOptions` array)
struct LabelAnnotationRendererView: View {
    @ObservedObject var viewContext: ViewContextController

    @State
    var selectedLabels = Set<String>()

    var labelType: String {
        viewContext.rendererDefinitionPropertyResolver.string("labelType") ?? "UNDEFINED"
    }
    
    var labelOptionIcons: [String] {
        viewContext.rendererDefinitionPropertyResolver.stringArray("labelOptionIcons")
    }

    var labelOptions: [LabelOption] {
        let icons = labelOptionIcons
        return viewContext.rendererDefinitionPropertyResolver.stringArray("labelOptions").indexed().map { label in
            LabelOption(
                labelID: label.element,
                text: label.element.titleCase(),
                icon: Image(systemName: icons[safe: label.index] ?? "tag")
            )
        }
    }
    

    func moveToPreviousItem() {
        selectedLabels = []
        guard viewContext.config.focusIndex > 0 else {
            moveToIndex(viewContext.items.endIndex - 1)
            return
        }
        moveToIndex(viewContext.config.focusIndex - 1)
    }

    func moveToNextItem() {
        guard viewContext.config.focusIndex < viewContext.items.endIndex - 1 else {
            moveToIndex(0)
            return
        }
        moveToIndex(viewContext.config.focusIndex + 1)
    }
    
    func moveToIndex(_ index: Int) {
        viewContext.config.focusIndex = index
        loadExistingAnnotation()
    }

    func loadExistingAnnotation() {
            selectedLabels = currentAnnotationLabels
    }

    func applyCurrentItem() {
        guard let currentItem = currentItem else { return }
        let labelsString = selectedLabels.joined(separator: ",")
        
        let annotationItem = currentAnnotation() ?? ItemRecord(type: "LabelAnnotation")
        try? annotationItem.setPropertyValue(name: "labels", value: .string(labelsString))
        
        let edge = ItemEdgeRecord(sourceUID: annotationItem.uid, name: "annotatedItem", targetUID: currentItem.uid)
        try? edge.save()
        
        moveToNextItem()
    }

    func currentAnnotation() -> ItemRecord? {
        self.currentItem?.reverseEdgeItem("annotatedItem")
    }

    var currentAnnotationLabels: Set<String> {
        currentAnnotation()?.propertyValue("labels")?.asString().map { Set($0.split(separator: ",").map(String.init)) } ?? []
    }

    var currentItem: ItemRecord? {
        viewContext.items[safe: viewContext.config.focusIndex]
    }

    var progressText: String? {
        guard !viewContext.items.isEmpty else { return nil }
        return "Item \(viewContext.config.focusIndex + 1) of \(viewContext.items.count)"
    }

    var enableBackButton: Bool {
        viewContext.config.focusIndex > 0
    }

    var enableSkipButton: Bool {
        viewContext.config.focusIndex < viewContext.items.endIndex - 1
    }

    @ViewBuilder
    var currentContent: some View {
        if let item = currentItem {
            viewContext.render(item: item)
        }
        else {
            Text("No items to label")
                .bold()
                .frame(maxWidth: .infinity, maxHeight: .infinity)
        }
    }

    var body: some View {
        LabelSelectionView(options: labelOptions,
                           selected: $selectedLabels,
                           enabled: currentItem != nil,
                           onBackPressed: moveToPreviousItem,
                           onCheckmarkPressed: applyCurrentItem,
                           onSkipPressed: moveToNextItem,
                           enableBackButton: enableBackButton,
                           enableCheckmarkButton: true,
                           enableSkipButton: enableSkipButton,
                           topText: progressText,
                           content: currentContent,
                           useScrollView: false)
            .onAppear {
                loadExistingAnnotation()
            }
    }
}

struct LabelOption {
    var labelID: String
    var text: String
    var icon: Image

    var id: String { labelID }
}

struct LabelSelectionView<Content: View>: View {
    var options: [LabelOption]
    

    @Binding var selected: Set<String>
    var enabled: Bool

    var onBackPressed: () -> Void
    var onCheckmarkPressed: () -> Void
    var onSkipPressed: () -> Void

    var enableBackButton: Bool
    var enableCheckmarkButton: Bool
    var enableSkipButton: Bool

    var topText: String?

    var content: Content
    var useScrollView: Bool = true

    var body: some View {
        VStack(spacing: 0) {
            topText.map { topText in
                Text(topText)
                    .font(Font.body.monospacedDigit())
                    .frame(maxWidth: .infinity)
                    .padding(.horizontal, 10)
                    .padding(.vertical, 5)
                    .background(Color(.secondarySystemBackground))
            }
            if topText != nil {
                Divider()
            }
            if useScrollView {
                GeometryReader { geometry in
                    ScrollView(.vertical) {
                        self.content
                            .frame(width: geometry.size.width)
                    }
                }
            }
            else {
                self.content
            }
            SwiftUI.List {
                ForEach(options, id: \.id) { option in
                    Button(action: {
                        if self.selected.remove(option.id) == nil {
                            self.selected.insert(option.id)
                        }
                    }) {
                        HStack {
                            option.icon
                            Text(option.text)
                        }
                        .frame(maxWidth: .infinity, alignment: .leading)
                        .if(self.selected.contains(option.id)) {
                            $0
                                .foregroundColor(.white)
                                .background(Color.blue.cornerRadius(4).padding(-5))
                        }
                        .contentShape(Rectangle())
                    }
                    .buttonStyle(PlainButtonStyle())
                }
            }
            .opacity(enabled ? 1 : 0.4)
            .frame(height: 220)
            .clipShape(RoundedCornerRectangle(radius: 20, corners: [.topLeft, .topRight]))
            .background(
                RoundedCornerRectangle(radius: 20, corners: [.topLeft, .topRight])
                    .fill(Color(.systemBackground))
                    .shadow(radius: 10)
            )
            Divider()
            HStack(spacing: 0) {
                Button(action: onBackPressed) {
                    Image(systemName: "arrow.uturn.left").font(.system(size: 20))
                        .padding(.horizontal, 20)
                        .contentShape(Rectangle())
                        .foregroundColor(enableBackButton ? Color.blue : Color.gray.opacity(0.5))
                }
                .disabled(!enableBackButton)
                Button(action: onCheckmarkPressed) {
                    Image(systemName: "checkmark").font(.system(size: 25))
                        .padding()
                        .frame(maxWidth: .infinity, maxHeight: .infinity)
                        .foregroundColor(.white)
                        .background(Color.green.opacity(enableCheckmarkButton ? 1 : 0.5))
                        .contentShape(Rectangle())
                }
                .disabled(!enableCheckmarkButton)
                Button(action: onSkipPressed) {
                    Text("Skip").font(.system(size: 20))
                        .padding(.horizontal, 20)
                        .contentShape(Rectangle())
                        .foregroundColor(enableSkipButton ? Color.blue : Color.gray.opacity(0.5))
                }
                .disabled(!enableSkipButton)
            }
            .opacity(enabled ? 1 : 0.4)
            .frame(height: 50)
            .background(Color(.secondarySystemBackground))
        }
        .disabled(!enabled)
    }
}
