//
//  NoteEditorRenderer.swift
//  Memri
//
//  Created by T Brennan on 22/1/21.
//

import Foundation
import Combine
import SwiftUI

/// The note editor renderer
/// This presents an editor for a single item
/// - for an item to be shown the CVU for ItemType > NoteEditor {...} must define both a `title` and `content` expression
/// - see the `Note.cvu` file for an example of this
/// - `File` items containing images can be referenced within the content HTML as `<img src="memriFile://fileUIDHere">`. The editor will look for a `file` edge from the item to a matching file, without this edge it will not be found (security measure)
struct NoteEditorRendererView: View {
    @EnvironmentObject var sceneController: SceneController
    @ObservedObject var viewContext: ViewContextController
    
    @State private var showingImagePicker: Bool = false
    @State private var showingImagePicker_shouldUseCamera: Bool = false
    @State private var imagePickerPromise: ((Result<URL?, Never>) -> Void)?
    
    var noteTitle: Binding<String?> {
        guard let item = viewContext.focusedItem,
              let resolver = viewContext.nodePropertyResolver(for: item),
              let binding: Binding<String?> = resolver.binding(for: "title")
        else {
            return .constant("")
        }
        return binding
    }
    
    var noteContent: Binding<String> {
        guard let item = viewContext.focusedItem,
              let resolver = viewContext.nodePropertyResolver(for: item),
              let binding = resolver.binding(for: "content", defaultValue: "")
        else {
            return .constant("")
        }
        return binding
    }
    
    func getEditorModel() -> MemriTextEditorModel {
        MemriTextEditorModel(
            title: noteTitle.wrappedValue,
            body: noteContent.wrappedValue
        )
    }

    func handleModelUpdate(_ newModel: MemriTextEditorModel) {
        if noteTitle.wrappedValue != newModel.title {
            noteTitle.wrappedValue = newModel.title
        }
        if noteContent.wrappedValue != newModel.body {
            noteContent.wrappedValue = newModel.body
        }
    }

    var body: some View {
        MemriTextEditor(
            model: getEditorModel,
            onModelUpdate: handleModelUpdate(_:),
            imageSelectionHandler: self,
            fileHandler: self,
            searchTerm: viewContext.searchString,
            isEditing: $sceneController.isInEditMode
        )
        .sheet(isPresented: $showingImagePicker) {
            if showingImagePicker_shouldUseCamera {
                CameraView { selectedImage in
                    onImagePickerFinished(image: selectedImage)
                }
            }
            else {
                ImagePickerView { selectedImage in
                    onImagePickerFinished(image: selectedImage)
                }
            }
        }
    }
}



extension NoteEditorRendererView: MemriTextEditorImageSelectionHandler {
    func presentImageSelectionUI(useCamera: Bool) -> AnyPublisher<URL?, Never> {
        Future<URL?, Never> { promise in
            self.imagePickerPromise = promise
            self.showingImagePicker_shouldUseCamera = useCamera
            self.showingImagePicker = true
        }.eraseToAnyPublisher()
    }
    
    func onImagePickerFinished(image: UIImage?) {
        var url: URL?
        defer { imagePickerPromise?(.success(url)) }

        guard let image = image, let note = viewContext.focusedItem else { return }
        let filename = UUID().uuidString

        do {
            guard let data = image.jpegData(compressionQuality: 0.8) else { return }
            try FileStorageController.writeData(data, toFileForUUID: filename)

            let file = ItemRecord(type: "File")
            let filenameProperty = ItemPropertyRecord(itemUID: file.uid, name: "filename", value: .string(filename))
            let edge = ItemEdgeRecord(sourceUID: note.uid, name: "file", targetUID: file.uid)
            
            try viewContext.databaseController.writeSync { db in
                try file.save()
                try filenameProperty.save()
                try edge.save()
            }
        } catch {
            print(error)
            return
        }

        // The value of url is passed to the completion promise (done by the defer statement)
        url = URL(string: "memriFile://\(filename)")
    }
}

extension NoteEditorRendererView: MemriWebViewFileHandler {
    func getFileData(forEditorURL url: URL) -> Data? {
        guard let fileUID = url.host,
              let note = viewContext.focusedItem,
              note.edgeItems("file").first(where: { $0.propertyValue("filename")?.asString() == fileUID }) != nil,
              let data = FileStorageController.getData(fromFileForUUID: fileUID)
        else { return nil }
        return data
    }
}
