//
// TimelineRenderer.swift
// Copyright © 2020 memri. All rights reserved.

import ASCollectionView
import SwiftUI

/// The timeline renderer
/// This presents the data in chronological order in a vertically scrolling `timeline`
struct TimelineRendererView: View {
    @EnvironmentObject var sceneController: SceneController
    @ObservedObject var viewContext: ViewContextController
   
    
    func generateModel() -> TimelineRendererModel {
        TimelineRendererModel(
            dataItems: viewContext.items,
            itemDateTimeResolver: { item -> Date? in
                viewContext.nodePropertyResolver(for: item)?.dateTime("dateTime") ?? item.dateModified
            },
            detailLevel: detailLevel,
            mostRecentFirst: mostRecentFirst
        )
    }
    
    var detailLevel: TimelineRendererModel.DetailLevel {
        viewContext.rendererDefinitionPropertyResolver.string("detailLevel").flatMap(TimelineRendererModel.DetailLevel.init) ?? .day
    }
    
    var mostRecentFirst: Bool {
        viewContext.rendererDefinitionPropertyResolver.bool("recentFirst", defaultValue: true)
    }

    let minSectionHeight: CGFloat = 40

    func sections(withModel model: TimelineRendererModel) -> [ASSection<Date>] {
        
        model.data.map { group in
            ASSection<Date>(id: group.date,
                            data: group.items,
                            selectionMode: .selectSingle { index in
                                guard let element = group.items[safe: index] else { return }
                                if element.isGroup {
                                    CVUAction_OpenView(
                                        renderer: "list",
                                        uids: Set(element.items.map { $0.uid })
                                    ).execute(sceneController: sceneController, context: viewContext.getCVUContext())
                                }
                                else if let item = element.items.first,
                                    let press = viewContext.nodePropertyResolver(for: item)?.action(key: "onPress") {
                                        press.execute(sceneController: sceneController, context: viewContext.getCVUContext(forItem: item))
                                }
                            }) { element, _ in
                self.renderElement(element)
                    .if(group.items.count < 2) { $0.frame(minHeight: self.minSectionHeight) }
            }
            .sectionHeader {
                header(withModel: model, for: group, calendarHelper: model.calendarHelper)
            }
            .sectionFooter {
                VStack {
                    Divider()
                }
            }
            .selfSizingConfig { (_) -> ASSelfSizingConfig? in
                .init(
                    selfSizeHorizontally: false,
                    selfSizeVertically: true,
                    canExceedCollectionWidth: false,
                    canExceedCollectionHeight: true
                )
            }
        }
    }

    @ViewBuilder
    func renderElement(_ element: TimelineElement) -> some View {
        if element.isGroup {
            TimelineItemView(icon: Image(systemName: "rectangle.stack"),
                             title: "\(element.items.count) \(element.itemType.titleCase())\(element.items.count != 1 ? "s" : "")",
                             backgroundColor: .gray)
                .frame(maxWidth: .infinity, alignment: .leading)
        }
        else if let item = element.items.first {
            self.viewContext.render(item: item)
                .frame(maxWidth: .infinity, alignment: .leading)
        }
    }

    var body: some View {
        let model = generateModel()
        ASCollectionView(sections: sections(withModel: model))
            .layout(layout(withModel: model))
            .alwaysBounceVertical()
    }

    var leadingInset: CGFloat {
        60
    }

    func layout(withModel model: TimelineRendererModel) -> ASCollectionLayout<Date> {
        ASCollectionLayout(scrollDirection: .vertical,
                           interSectionSpacing: 0) { () -> ASCollectionLayoutSection in
            ASCollectionLayoutSection { _ -> NSCollectionLayoutSection in
                let hasFullWidthHeader: Bool = model.detailLevel == .year

                let itemLayoutSize = NSCollectionLayoutSize(
                    widthDimension: .fractionalWidth(1.0),
                    heightDimension: .estimated(20)
                )
                let groupSize = NSCollectionLayoutSize(
                    widthDimension: .fractionalWidth(1.0),
                    heightDimension: .estimated(20)
                )

                let item = NSCollectionLayoutItem(layoutSize: itemLayoutSize)
                let group = NSCollectionLayoutGroup.horizontal(
                    layoutSize: groupSize,
                    subitem: item,
                    count: 1
                )

                let section = NSCollectionLayoutSection(group: group)

                section.contentInsets = .init(
                    top: 8,
                    leading: hasFullWidthHeader ? 10 : self.leadingInset + 5,
                    bottom: 8,
                    trailing: 10
                )
                section.interGroupSpacing = 10
                section.visibleItemsInvalidationHandler = { _, _, _ in
                    // If this isn't defined, there is a bug in UICVCompositional Layout that will fail to update sizes of cells
                }

                var headerSupplementary: NSCollectionLayoutBoundarySupplementaryItem
                if hasFullWidthHeader {
                    let supplementarySize = NSCollectionLayoutSize(
                        widthDimension: .fractionalWidth(1.0),
                        heightDimension: .absolute(35)
                    )
                    headerSupplementary = NSCollectionLayoutBoundarySupplementaryItem(
                        layoutSize: supplementarySize,
                        elementKind: UICollectionView.elementKindSectionHeader,
                        alignment: .topLeading
                    )
                    headerSupplementary.extendsBoundary = true
                    headerSupplementary.pinToVisibleBounds = false
                }
                else {
                    let supplementarySize = NSCollectionLayoutSize(
                        widthDimension: .absolute(self.leadingInset),
                        heightDimension: .absolute(self.minSectionHeight + 16)
                    )
                    headerSupplementary = NSCollectionLayoutBoundarySupplementaryItem(
                        layoutSize: supplementarySize,
                        elementKind: UICollectionView.elementKindSectionHeader,
                        alignment: .topLeading
                    )
                    headerSupplementary.extendsBoundary = false
                    headerSupplementary.pinToVisibleBounds = true
                }

                let footerSupplementary = NSCollectionLayoutBoundarySupplementaryItem(
                    layoutSize: NSCollectionLayoutSize(
                        widthDimension: .fractionalWidth(1.0),
                        heightDimension: .absolute(1)
                    ),
                    elementKind: UICollectionView.elementKindSectionFooter,
                    alignment: .bottom
                )

                section.supplementariesFollowContentInsets = false
                section.boundarySupplementaryItems = [headerSupplementary, footerSupplementary]
                return section
            }
        }
    }
}

extension TimelineRendererView {
    // TODO: Clean up this function. Should probably define for each `DetailLevel` individually
    func header(withModel model: TimelineRendererModel, for group: TimelineGroup, calendarHelper: CalendarHelper) -> some View {
        let matchesNow = calendarHelper.isSameAsNow(
            group.date,
            byComponents: model.detailLevel.relevantComponents
        )

        let flipOrder: Bool = {
            switch model.detailLevel {
            case .hour: return true
            default:
                return false
            }
        }()

        let alignment: HorizontalAlignment = {
            switch model.detailLevel {
            case .year: return .leading
            case .day: return .center
            default: return .trailing
            }
        }()

        let largeString: String? = {
            switch model.detailLevel {
            case .hour:
                if group.isStartOf.contains(.day) {
                    let format = DateFormatter()
                    format.dateFormat = "dd/MM"
                    return format.string(from: group.date)
                }
            case .day:
                let format = DateFormatter()
                format.dateFormat = "d"
                return format.string(from: group.date)
            case .week:
                let format = DateFormatter()
                format.dateFormat = "ww"
                return format.string(from: group.date)
            case .month:
                let format = DateFormatter()
                format.dateFormat = "MMM"
                return format.string(from: group.date)
            case .year:
                let format = DateFormatter()
                format.dateFormat = "YYYY"
                return format.string(from: group.date)
            }
            return nil
        }()

        let smallString: String? = {
            switch model.detailLevel {
            case .hour:
                let format = DateFormatter()
                format.dateFormat = "h a"
                return format.string(from: group.date)
            case .day:
                let format = DateFormatter()
                format.dateFormat = "MMM" // group.isStartOf.contains(.year) ? "MMM YY" : "MMM"
                return format.string(from: group.date)
            case .week:
                return "Week"
            case .month:
                if group.isStartOf.contains(.year) {
                    let format = DateFormatter()
                    format.dateFormat = "YYYY"
                    return format.string(from: group.date)
                }
            default: break
            }
            return nil
        }()

        let small: some View = {
            smallString.map { string in
                Text(string)
                    .font(Font.system(size: 14))
                    .foregroundColor(matchesNow ? Color.red : Color(.secondaryLabel))
                    .fixedSize()
            }
        }()

        return VStack(alignment: alignment, spacing: 0) {
            if !flipOrder {
                small
            }
            largeString.map { string in
                Text(string)
                    .font(Font.system(size: 20))
                    .lineLimit(1)
                    .minimumScaleFactor(0.6)
                    .foregroundColor(matchesNow ? (useFillToIndicateNow(model: model) ? Color.white : .red) :
                        Color(.label))
                    .padding(.vertical, matchesNow ? 3 : 0)
                    .background(
                        Circle().fill((useFillToIndicateNow(model: model) && matchesNow) ? Color.red : .clear)
                            .frame(minWidth: 30, minHeight: 30)
                    )
                    .fixedSize()
            }
            if flipOrder {
                small
            }
            Spacer(minLength: 0)
        }
        .padding(8)
        .frame(maxWidth: .infinity, alignment: Alignment(horizontal: alignment, vertical: .top))
    }

    func useFillToIndicateNow(model: TimelineRendererModel) -> Bool {
        switch model.detailLevel {
        case .day:
            return true
        default:
            return false
        }
    }
}
