//
// TimelineRendererModel.swift
// Copyright © 2020 memri. All rights reserved.

import Foundation

/// The model used to group and display data on the timeline
struct TimelineRendererModel {
    var data: [TimelineGroup]
    var detailLevel: DetailLevel
    var mostRecentFirst: Bool

    var itemDateTimeResolver: (ItemRecord) -> Date?

    let calendarHelper = CalendarHelper()

    init(
        dataItems: [ItemRecord],
        itemDateTimeResolver: @escaping ((ItemRecord) -> Date?),
        detailLevel: DetailLevel,
        mostRecentFirst: Bool
    ) {
        self.detailLevel = detailLevel
        self.mostRecentFirst = mostRecentFirst
        self.itemDateTimeResolver = itemDateTimeResolver

        data = TimelineRendererModel.group(
            dataItems,
            itemDateTimeResolver: itemDateTimeResolver,
            level: detailLevel,
            mostRecentFirst: mostRecentFirst
        )
    }

    static func group(
        _ data: [ItemRecord],
        itemDateTimeResolver: (ItemRecord) -> Date?,
        level: TimelineRendererModel.DetailLevel,
        mostRecentFirst: Bool,
        collapseByTypeWhereCountGreaterThan maxCount: Int = 2
    ) -> [TimelineGroup] {
        let dataPairs: [(ItemRecord, Date)] = data.compactMap { dataItem in
            guard let date = itemDateTimeResolver(dataItem) else { return nil }
            return (dataItem, date)
        }
        let groupedByDetailLevel = Dictionary(grouping: dataPairs) { (dataItem, date) -> Date? in
            let components = Calendar.current.dateComponents(level.relevantComponents, from: date)
            guard let day = Calendar.current.date(from: components) else { return nil }

            return day
        }
        var sortedGroups = groupedByDetailLevel.compactMap { (date, itemsWithDate) -> TimelineGroup? in
            guard let date = date else { return nil }
            return TimelineGroup(
                date: date,
                items: group(itemsWithDate, mostRecentFirst: mostRecentFirst, collapseByTypeWhereCountGreaterThan: maxCount)
            )
        }.sorted(by: { mostRecentFirst ? ($0.date > $1.date) : ($0.date < $1.date) })

        guard !sortedGroups.isEmpty else { return [] }

        let largerComponents = level.largerComponents
        sortedGroups[0].isStartOf = largerComponents
        for index in sortedGroups.indices.dropFirst() {
            let difference = Calendar.current.dateComponents(
                largerComponents,
                from: sortedGroups[index - 1].date,
                to: sortedGroups[index].date
            )
            sortedGroups[index].isStartOf = largerComponents
                .filter { (difference.value(for: $0).map { mostRecentFirst ? -$0 : $0 } ?? 0) > 0 }
        }

        return sortedGroups
    }

    static func group(
        _ data: [(ItemRecord, Date)],
        mostRecentFirst: Bool,
        collapseByTypeWhereCountGreaterThan maxCount: Int
    ) -> [TimelineElement] {
        let groupedByType = Dictionary(grouping: data) { dataItem -> String in
            dataItem.0.type
        }
        let groups = groupedByType.flatMap { (type, itemsWithDate) -> [TimelineElement] in
            if itemsWithDate.count > maxCount {
                return [TimelineElement(itemType: type, index: 0, items: itemsWithDate.map { $0.0 }, date: itemsWithDate.max(by: { mostRecentFirst ? ($0.1 > $1.1) : ($0.1 < $1.1) })?.1 ?? Date())]
            } else {
                return itemsWithDate.indexed()
                    .map { TimelineElement(itemType: type, index: $0.index, items: [$0.element.0], date: $0.1) }
            }
        }.sorted(by: { mostRecentFirst ? ($0.date > $1.date) : ($0.date < $1.date) })
        return groups
    }
}

struct TimelineGroup: Equatable {
    var date: Date
    var items: [TimelineElement]

    // Used to store whether this is the first entry in year/month/day etc (for use in rendering supplementaries)
    var isStartOf: Set<Calendar.Component> = []
}

struct TimelineElement: Equatable {
    var itemType: String
    var index: Int
    var items: [ItemRecord]
    var date: Date

    var isGroup: Bool { items.first != items.last }
}

extension TimelineElement: Hashable, Identifiable {
    func hash(into hasher: inout Hasher) {
        hasher.combine(itemType)
        hasher.combine(index)
    }

    static func == (lhs: TimelineElement, rhs: TimelineElement) -> Bool {
        lhs.itemType == rhs.itemType && lhs.index == rhs.index
    }

    var id: Int {
        hashValue
    }
}

extension TimelineRendererModel {
    enum DetailLevel: String, CaseIterable {
        case year
        case month
        case week
        case day
        case hour

        var relevantComponents: Set<Calendar.Component> {
            switch self {
            case .year: return [.year]
            case .month: return [.year, .month]
            case .week: return [
                    .yearForWeekOfYear,
                    .weekOfYear,
                ] // Note yearForWeekOfYear is used to correctly account for weeks crossing the new year
            case .day: return [.year, .month, .day]
            case .hour: return [.year, .month, .day, .hour]
            }
        }

        var largerComponents: Set<Calendar.Component> {
            switch self {
            case .year: return []
            case .month: return [.year]
            case .week: return [.year] // Note yearForWeekOfYear is used to correctly account for weeks crossing the new year
            case .day: return [.year, .month]
            case .hour: return [.year, .month, .day]
            }
        }
    }
}
