//
// SetupWizard.swift
// Copyright © 2020 memri. All rights reserved.

import Foundation
import SwiftUI

/// This screen is displayed on the very first launch of the app. It allows the user to configure a connection to a pod (or a local demo)
struct SetupScreenView: View {
    @ObservedObject var appController: AppController = AppController.shared
    @State var model = SetupScreen_Model()
    @State var showingNewPodWarning: Bool = false
    
    var body: some View {
        ZStack {
            NavigationView {
                VStack(spacing: 10) {
                    VStack(spacing: -10) {
                        Text("Welcome to")
                            .font(Font.system(size: 20, weight: .light))
                        Text("memri")
                            .font(.system(size: 60, weight: .bold, design: .default))
                            .foregroundColor(.purple)
                    }
                    Text("A place where your data belongs to you.")
                        .multilineTextAlignment(.center)
                        .foregroundColor(Color(.label))
                    Spacer().frame(height: 30)
                    VStack(alignment: .leading, spacing: 6) {
                        Text("Have a memri pod?")
                        NavigationLink(
                            destination: podSetup,
                            label: {
                                Text("Connect to pod")
                                                            .font(.headline)
                                                            .foregroundColor(.white)
                                                            .frame(maxWidth: .infinity, minHeight: 50)
                                                            .background(Color.green.cornerRadius(10))
                            })
                            .buttonStyle(PlainButtonStyle())
                    }
                    VStack(alignment: .leading, spacing: 6) {
                        Text("Just want to try the app?")
                        Button(action: onLocalDemoPressed) {
                                Text("Let me try the app without a pod")
                                                            .frame(maxWidth: .infinity, minHeight: 50)
                                    .background(Color(.tertiarySystemBackground).cornerRadius(10))
                            }
                            .buttonStyle(PlainButtonStyle())
                    }
                    .padding(.top, 10)
                }
                .padding()
                .frame(maxWidth: .infinity, maxHeight: .infinity)
                .background(Color(.secondarySystemBackground).edgesIgnoringSafeArea(.all))
            }
            .colorScheme(.dark)
            .navigationViewStyle(StackNavigationViewStyle())
            .alert(isPresented: $showingNewPodWarning, content: {
                newPodWarning
            })
            if model.state == .loading {
                Color.black.opacity(0.75)
                VStack(spacing: 10) {
                    ActivityIndicatorView(style: .large, color: .white)
                    Text("Setup in progress...").foregroundColor(.white)
                }
            }
        }
    }
    
    var captionFont: Font {
        Font.caption.bold()
    }
    var podSetup: some View {
        Form {
            Section(header:
                        Text("Connection Details")
                    ) {
                VStack(alignment: .leading, spacing: 2) {
                    Text("Pod URL").font(captionFont)
                    MemriTextField(value: $model.podURL, placeholder: model.defaultPodURL)
                }
            }
            
            Section(header: Text("Pod setup"),
                    footer: connectButton.showIf(model.setupAsNewPod)
            ) {
                Toggle("Set up as new pod?", isOn: $model.setupAsNewPod)
                Text("If enabled this will create new authentication keys and install the demo data.")
                    .font(.caption)
            }
            if !model.setupAsNewPod {
                Section(header: Text("Authentication"),
                        footer: connectButton.showIf(!model.setupAsNewPod)
                ) {
                    VStack(alignment: .leading, spacing: 2) {
                        Text("Public Key").font(captionFont)
                        MemriTextField(value: $model.podPublicKey, placeholder: "publickey")
                    }
                    VStack(alignment: .leading, spacing: 2) {
                        Text("Private Key").font(captionFont)
                        MemriTextField(value: $model.podPrivateKey, placeholder: "privatekey", secureMode: !Constants.isSimulator)
                    }
                    VStack(alignment: .leading, spacing: 2) {
                        Text("Database Key").font(captionFont)
                        MemriTextField(value: $model.podDatabaseKey, placeholder: "databasekey", secureMode: !Constants.isSimulator)
                    }
                }
            }
        }
        .navigationTitle("Pod Setup")
    }
    
    var connectButton: some View {
        VStack(spacing: 10) {
            if case let .error(errorString) = model.state {
                Text("Error connecting to pod: \(errorString)")
                    .multilineTextAlignment(.center)
                    .lineLimit(3)
                    .foregroundColor(.red)
            }
            if !model.isValidToProceedToConnect {
                Text("You must provide the public, private, and database key to connect to an existing pod.")
                    .foregroundColor(.red)
            }
            Button(action: onConnectPressed) {
                Text("Connect to pod")
                    .font(.headline)
                    .foregroundColor(.white)
                    .frame(maxWidth: .infinity, minHeight: 50)
                    .background(Color.green.cornerRadius(10))
            }.buttonStyle(PlainButtonStyle())
            .disabled(!model.isValidToProceedToConnect)
        }
        .padding(.top)
    }
    
    func onConnectPressed() {
        if model.setupAsNewPod {
            showingNewPodWarning = true
        } else {
            handleSetup(localOnly: false)
        }
    }
    
    func onLocalDemoPressed() {
        handleSetup(localOnly: true)
    }
    
    func onAcknowledgedNewPodWarning() {
        handleSetup(localOnly: false)
    }
    
    func handleSetup(localOnly: Bool) {
        model.state = .loading
        
        func handleCompletion(error: Error?) {
            if let error = error {
                model.state = .error("\(error)")
            } else {
                model.state = .idle
            }
        }
        
        guard let config = model.getSetupConfig(localOnly: localOnly) else {
            handleCompletion(error: nil)
            return
        }
        appController.setupApp(config: config, onCompletion: handleCompletion)
    }
    
    
    var newPodWarning: Alert {
        Alert(title: Text("Set up new pod"),
              message: Text("Are you sure you want to install demo data to your pod?"),
              primaryButton: .default(Text("Set up as new pod"), action: onAcknowledgedNewPodWarning),
              secondaryButton: .cancel())
    }
}

struct SetupScreen_Previews: PreviewProvider {
    static var previews: some View {
        SetupScreenView()
    }
}
