//
// ExprInterpreterTests.swift
// Copyright © 2020 memri. All rights reserved.

@testable import Memri
import XCTest
import GRDB


class ExprInterpreterTests: XCTestCase {
    override func setUpWithError() throws {}

    override func tearDownWithError() throws {}

    private func parse(_ snippet: String, stringMode: Bool = false) throws -> CVUValue.ExpressionNode {
        let lexer = CVUExpressionLexer(input: snippet, startInStringMode: stringMode)
        let tokens = try lexer.tokenize()
        let parser = CVUExpressionParser(tokens)
        return try parser.parse()
    }
    
    let databaseController = DatabaseController()
    let lookupController = CVULookupController(mockMode: .init(bool: true, string: "Memri", double: 10))
    
    private func interpretAsBool(_ expr: CVUValue.ExpressionNode) throws -> Bool? {
        lookupController.resolve(value: .expression(expr), context: CVUContext(), db: databaseController)
    }
    
    private func interpretAsDouble(_ expr: CVUValue.ExpressionNode) throws -> Double? {
        lookupController.resolve(value: .expression(expr), context: CVUContext(), db: databaseController)
    }
    
    private func interpretAsString(_ expr: CVUValue.ExpressionNode) throws -> String? {
        lookupController.resolve(value: .expression(expr), context: CVUContext(), db: databaseController)
    }

    func testArithmeticOperators() throws {
        let snippet = "(5 + 10 * 4 - 3 / 10) / 10"

        let expr = try parse(snippet)
        let result = try interpretAsDouble(expr)

        XCTAssertEqual(result, (5 + 10 * 4 - 3 / 10) / 10)
    }

    func testAnd() throws {
        let snippet = "true and false"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)
        
        XCTAssertEqual(result, false)
    }

    func testOr() throws {
        let snippet = "true or false"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)

        XCTAssertEqual(result, true)
    }

    func testEqualsTrue() throws {
        let snippet = "1 = '1'"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)

        XCTAssertEqual(result, true)
    }

    func testEqualsFalse() throws {
        let snippet = "1 = 2"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)

        XCTAssertEqual(result, false)
    }

    func testOrValue() throws {
        let snippet = "10 or 0"

        let expr = try parse(snippet)
        let result = try interpretAsDouble(expr)

        XCTAssertEqual(result, 10)
    }

    func testSimpleCondition() throws {
        let snippet = "true ? 'yes' : 'no'"

        let expr = try parse(snippet)
        let result = try interpretAsString(expr)

        XCTAssertEqual(result, "yes")
    }

    func testMultiCondition() throws {
        let snippet = "true ? false and true ? -1 : false or true ? 'yes' : 'no' : -1"

        let expr = try parse(snippet)
        let result = try interpretAsString(expr)

        XCTAssertEqual(result, "yes")
    }

    func testConditionEquals() throws {
        let snippet = "true = false"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)

        XCTAssertEqual(result, false)
    }

    func testConditionNotEquals() throws {
        let snippet = "true != false"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)

        XCTAssertEqual(result, true)
    }

    func testConditionGreaterThan() throws {
        let snippet = "5 > 10"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)

        XCTAssertEqual(result, false)
    }

    func testConditionGreaterThanOrEqual() throws {
        let snippet = "5 >= 5"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)

        XCTAssertEqual(result, true)
    }

    func testConditionLessThan() throws {
        let snippet = "5 < 10"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)

        XCTAssertEqual(result, true)
    }

    func testConditionLessThanOrEqual() throws {
        let snippet = "5 <= 5"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)

        XCTAssertEqual(result, true)
    }

//    func testLookup() throws {
//        let snippet = ".bar and bar.foo(10) and bar[foo = 10] or shouldNeverGetHere"
//
//        var results = [CVUValue.ExpressionNode]()
//
//        let tree = try parse(snippet)
//        let interpreter = ExprInterpreter(tree,
//                                          { lookup, _ in
//                                              results.append(lookup)
//                                              return true
//                                          },
//                                          { lookup, args, _ in
//                                              results.append(lookup)
//                                              XCTAssertEqual(args[0] as! Double, 10)
//                                              return true
//                                          })
//        let result = try interpreter.execute()
//
//        XCTAssertEqual(results.count, 3)
//
//        XCTAssertEqual(
//            results[0].description,
//            "LookupNode([VariableNode(@@DEFAULT@@, type:propertyOrItem, list:single), VariableNode(bar, type:propertyOrItem, list:single)])"
//        )
//        XCTAssertEqual(
//            results[1].description,
//            "LookupNode([VariableNode(bar, type:propertyOrItem, list:single), VariableNode(foo, type:propertyOrItem, list:single)])"
//        )
//        XCTAssertEqual(
//            results[2].description,
//            "LookupNode([VariableNode(bar, type:propertyOrItem, list:list), LookupNode([BinaryOpNode(ConditionEquals, lhs: LookupNode([VariableNode(foo, type:propertyOrItem, list:single)]), rhs: NumberNode(10.0))])])"
//        )
//
//        XCTAssertEqual(result as! Bool, true)
//    }

    func testMinusPlusModifier() throws {
        let snippet = "-5 + -(5+10) - +'5'"

        let expr = try parse(snippet)
        let result = try interpretAsDouble(expr)

        XCTAssertEqual(result, -25)
    }

    func testNegation() throws {
        let snippet = "!true"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)

        XCTAssertEqual(result, false)
    }
//
//    func testNegationWithLookup() throws {
//        let snippet = "!.label"
//
//        let tree = try parse(snippet)
//        let interpreter = ExprInterpreter(
//            tree,
//            { _, _ in
//                nil
//            },
//            { _, _, _ in
//                true
//            }
//        )
//        let result = try interpreter.execute()
//
//        XCTAssertEqual(result as! Bool, true)
//    }

    func testStringEscaping() throws {
        let snippet = "'asdadsasd\\'asdasd'"

        let expr = try parse(snippet)
        let result = try interpretAsString(expr)

        XCTAssertEqual(result, "asdadsasd'asdasd")
    }

    func testTypeConversionToNumber() throws {
        let snippet = "5 + '10.34' + true"

        let expr = try parse(snippet)
        let result = try interpretAsDouble(expr)

        XCTAssertEqual(result, 16.34)
    }

    func testNanStringToInt() throws {
        let snippet = "+'asdasd'"

        let expr = try parse(snippet)
        let result = try interpretAsDouble(expr)

        XCTAssertTrue(result == nil)
    }

    func testTypeConversionToBool() throws {
        let snippet = "0 ? -1 : 1 ? '' ? -1 : 'yes' : -1"

        
        let expr = try parse(snippet)
        let result = try interpretAsString(expr)

        XCTAssertEqual(result, "yes")
    }

    func testTypeConversionStringToBool() throws {
        let snippet = "''"

        let expr = try parse(snippet)
        let result = try interpretAsBool(expr)

        XCTAssertEqual(result, false)
    }

    func testStringModeStartWithString() throws {
        let snippet = "Hello {fetchName()}!"
        
        let expr = try parse(snippet, stringMode: true)
        let result = try interpretAsString(expr)

        XCTAssertEqual(result, "Hello Memri!")
    }

    func testStringModeStartWithExpression() throws {
        let snippet = "{fetchName()} Hello"
        
        let expr = try parse(snippet, stringMode: true)
        let result = try interpretAsString(expr)

        XCTAssertEqual(result, "Memri Hello")
    }
//
//    func testExample() throws {
//        let snippet = """
//        !(test + -5.63537) or 4/3 ? variable.func() : me.address[primary = true].country ? ((4+5 * 10) + test[10]) : 'asdads\\'asdad' + ''
//        """
//
//        let tree = try parse(snippet)
//        let interpreter = ExprInterpreter(tree,
//                                          { _, _ in 10 },
//                                          { _, _, _ in 20 })
//        let result = try interpreter.execute()
//
//        XCTAssertEqual(result as! Int, 20)
//    }
//
//    func testErrorLookupFailure() throws {
//        let snippet = ".bar"
//
//        let tree = try parse(snippet)
//        let interpreter = ExprInterpreter(tree,
//                                          { _, _ in throw "Undefined variable" },
//                                          { _, _, _ in 1 })
//
//        do {
//            _ = try interpreter.execute()
//        }
//        catch {
//            XCTAssertEqual(error as! String, "Undefined variable")
//            return
//        }
//
//        XCTFail()
//    }

}
