//
//  MemriPodAPITests.swift
//  MemriPodAPITests
//
//  Created by T Brennan on 19/12/20.
//

import XCTest
import Combine
@testable import Memri

/// This test suite requires a running pod in order to execute successfully.
class MemriPodAPITests: XCTestCase {


    
    /// This connection config is used to connect to the pod for the tests. You can change url scheme/path/port etc here
    let connectionConfig = PodAPIConnectionDetails(ownerKey: "TESTPOD",
                                                   databaseKey: "DBKEY")
    
    /// These are used to create test items in the pod for use in later tests
    let noteUID = StringUUID()
    let indexerUID = StringUUID()
    let importerUID = StringUUID()
    let fileUID = StringUUID()
    
    let schema = try! Schema.loadFromFile()
    
    private var testCancellables: Set<AnyCancellable> = []
    
    
    func setupPodForTesting() throws {
        let testNoteItem: [String: AnyEncodable] = {
            let testItem = ItemRecord(uid: noteUID, type: "Note")
            let testItemProperties: [ItemPropertyRecord] = [
                ItemPropertyRecord(itemUID: testItem.uid, name: "content", value: .string("Test note content")),
                ItemPropertyRecord(itemUID: testItem.uid, name: "title", value: .string("Test note title"))
            ]
            return testItem._syncDict(properties: testItemProperties, schema: schema)
        }()
        
        let testImporterItem: [String: AnyEncodable] = {
            let testItem = ItemRecord(uid: importerUID, type: "Importer")
            let testItemProperties: [ItemPropertyRecord] = [
                ItemPropertyRecord(itemUID: testItem.uid, name: "repository", value: .string("importerRepo"))
            ]
            return testItem._syncDict(properties: testItemProperties, schema: schema)
        }()
        
        let testIndexerItem: [String: AnyEncodable] = {
            let testItem = ItemRecord(uid: indexerUID, type: "Indexer")
            let testItemProperties: [ItemPropertyRecord] = [
                ItemPropertyRecord(itemUID: testItem.uid, name: "repository", value: .string("indexerRepo"))
            ]
            return testItem._syncDict(properties: testItemProperties, schema: schema)
        }()
        
        let testFileItem: [String: AnyEncodable] = {
            let testItem = ItemRecord(uid: fileUID, type: "File")
            let testItemProperties: [ItemPropertyRecord] = [
                ItemPropertyRecord(itemUID: testItem.uid,
                                                                               name: "sha256",
                                                                               value: .string(testFileSHA))
            ]
            return testItem._syncDict(properties: testItemProperties, schema: schema)
        }()
        
        let testEdge: [String: AnyEncodable] = {
            let testEdge = ItemEdgeRecord(sourceUID: noteUID, name: "file", targetUID: fileUID)
            return testEdge._syncDict()
        }()
        
        
        let bulkAction = PodAPIPayload.BulkAction(
            createItems: [testNoteItem, testImporterItem, testIndexerItem, testFileItem],
            updateItems: [],
            deleteItems: [],
            createEdges: [testEdge],
            updateEdges: [],
            deleteEdges: [])
        
        let request = PodAPIStandardRequest.bulkAction(payload: bulkAction)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
        }
    }
    
    override func setUpWithError() throws {
        try setupPodForTesting()
    }

    override func tearDownWithError() throws {
        // Put teardown code here. This method is called after the invocation of each test method in the class.
    }
    
    
    /// This function is used to await a network request to the pod before continuing
    /// It accepts a throwing function handling the response
    /// **Note:** in its current form this function assumes only a single value is returned (ie. one-off publishers)
    func await<T: Publisher>(publisher: T, handleValue: @escaping (T.Output) throws -> Void) {
        let e = expectation(description: "PodResponse")
    
        publisher.sink { (completion) in
            switch completion {
            case .failure(let error):
                XCTFail(error.localizedDescription)
            case .finished:
                break
            }
        } receiveValue: { (value) in
            do {
                try handleValue(value)
            }
            catch {
                XCTFail(error.localizedDescription)
            }
            e.fulfill()
        }.store(in: &testCancellables)

        waitForExpectations(timeout: 10.0, handler: nil)
    }
    
    

    func testRunImporter() throws {
        let request = PodAPIStandardRequest.runImporter(importerUID: importerUID, connectionConfig: connectionConfig)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
        }
    }
    
    func testRunIndexer() throws {
        let request = PodAPIStandardRequest.runIndexer(indexerUID: indexerUID, connectionConfig: connectionConfig)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
        }
    }
    
    func testGetItemWithEdges() throws {
        let request = PodAPIStandardRequest.getItemWithEdges(itemUID: noteUID)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
            let items = try response.result.get() as! [[String: Any]]
            XCTAssert(items.count == 1)
        }
    }
    
    
    func testGetItemsWithEdges() throws {
        let request = PodAPIStandardRequest.getItemsWithEdges(itemUIDs: [noteUID, indexerUID, importerUID])
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
            let items = try response.result.get() as! [[String: Any]]
            XCTAssert(!items.isEmpty)
        }
    }
    
    func testSearchByFields() throws {
        let request = PodAPIStandardRequest.searchByFields(itemType: "Note")
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
            let items = try response.result.get() as! [[String: Any]]
            XCTAssert(!items.isEmpty)
        }
    }
    
    func testGetAllItems() throws {
        let request = PodAPIStandardRequest.getAllItems()
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
            let items = try response.result.get() as! [[String: Any]]
            XCTAssert(!items.isEmpty)
        }
    }
    
    func testDownloadFile() throws {
        let request = PodAPIDownloadRequest.downloadFile(fileSHAHash: "2d05f86fec5b76dc7adcb19657885dc4844b7620edb0b5832b857a544a8eb08a", fileUID: "testFile")
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            let localFileURL = try response.result.get()
            let _ = localFileURL
        }
    }
    
    func testCreateItem() throws {
        let item = ItemRecord(type: "Note")
        let properties = [
            ItemPropertyRecord(itemUID: item.uid, name: "content", value: .string("Test note content")),
            ItemPropertyRecord(itemUID: item.uid, name: "title", value: .string("Test note title"))
        ]
        let syncDict = item._syncDict(properties: properties, schema: schema)
        let request = PodAPIStandardRequest.createItem(syncDict: syncDict)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
        }
    }
    
    let testFileSHA = "4156ae4d5c719f5adba859be68e5c4541058863dde6e76c93a6a0a3f0f9a361b"
    func testUploadFile() throws {
        // Make an upload request
        let url = Bundle.main.url(forResource: "schema", withExtension: "json")!
        let request = try PodAPIUploadRequest.uploadFile(fileURL: url, connectionConfig: connectionConfig)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            do {
                try response.throwIfError()
            }
            catch {
                // If file already exists, ignore this error -> otherwise throw
                if error.localizedDescription != "POD RESPONDED WITH ERROR: File already exists" {
                 throw error
                }
            }
        }
    }
    
    
    func testBulkAction() throws {
        let testCreateItem = ItemRecord(type: "Note")
        let testCreateItemProperties = [
            ItemPropertyRecord(itemUID: testCreateItem.uid, name: "content", value: .string("Test note content")),
            ItemPropertyRecord(itemUID: testCreateItem.uid, name: "title", value: .string("Test note title"))
        ]
        let testCreateItemSyncDict = testCreateItem._syncDict(properties: testCreateItemProperties, schema: schema)
        
        let testUpdateItem = ItemRecord(type: "Note")
        let testUpdateItemProperties = [
            ItemPropertyRecord(itemUID: testUpdateItem.uid, name: "content", value: .string("Updated note content")),
            ItemPropertyRecord(itemUID: testUpdateItem.uid, name: "title", value: .string("Updated note title"))
        ]
        let testUpdateItemSyncDict = testUpdateItem._syncDict(properties: testUpdateItemProperties, schema: schema)
        
        
        let bulkAction = PodAPIPayload.BulkAction(
            createItems: [testCreateItemSyncDict],
            updateItems: [testUpdateItemSyncDict],
            deleteItems: [],
            createEdges: [],
            updateEdges: [],
            deleteEdges: [])
        
        let request = PodAPIStandardRequest.bulkAction(payload: bulkAction)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
        }
    }
    
    func testSync() throws {
        let database = DatabaseController(databaseName: "podAPITestDB")
        try database.setupDatabase()
        try DemoData.importDemoData(databaseController: database, throwIfAgainstSchema: false)
        
        let syncController = SyncController(databaseController: database)
        let bulkAction = try syncController.makeSyncUploadData(maxItems: 10000)
        
        
        let request = PodAPIStandardRequest.bulkAction(payload: bulkAction)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
        }
    }
}

import Alamofire
extension DataResponse {
    func throwIfError() throws {
        switch result {
        case .success:
            break
        case let .failure(error):
            if let data = data,
               let responseString = String(data: data, encoding: .utf8) {
                throw StringError(description: "POD RESPONDED WITH ERROR: \(responseString)")
            } else {
                throw error
            }
        }
    }
}
